# coding: utf-8

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=line-too-long
"""SuperGLUEBenchmark corpora."""

__all__ = ['SuperGlueRTE', 'SuperGlueCB', 'SuperGlueWSC', 'SuperGlueWiC',
           'SuperGlueCOPA', 'SuperGlueMultiRC', 'SuperGlueBoolQ',
           'SuperGlueReCoRD', 'SuperGlueAXb', 'SuperGlueAXg']

import zipfile
import os
import re

from mxnet.gluon.utils import download, check_sha1, _get_repo_file_url

from .dataset import _JsonlDataset
from .registry import register
from ..base import get_home_dir


class _SuperGlueDataset(_JsonlDataset):
    def __init__(self, root, data_file):
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        segment, zip_hash, data_hash = data_file
        self._root = root
        filename = os.path.join(self._root, '%s.jsonl' % segment)
        self._get_data(segment, zip_hash, data_hash, filename)
        super(_SuperGlueDataset, self).__init__(filename)

    def _get_data(self, segment, zip_hash, data_hash, filename):
        data_filename = '%s-%s.zip' % (segment, data_hash[:8])
        if not os.path.exists(filename) or not check_sha1(filename, data_hash):
            download(_get_repo_file_url(self._repo_dir(), data_filename),
                     path=self._root, sha1_hash=zip_hash)
            # unzip
            downloaded_path = os.path.join(self._root, data_filename)
            with zipfile.ZipFile(downloaded_path, 'r') as zf:
                # skip dir structures in the zip
                for zip_info in zf.infolist():
                    if zip_info.filename[-1] == '/':
                        continue
                    zip_info.filename = os.path.basename(zip_info.filename)
                    zf.extract(zip_info, self._root)

    def _repo_dir(self):
        raise NotImplementedError


@register(segment=['train', 'val', 'test'])
class SuperGlueRTE(_SuperGlueDataset):
    """The Recognizing Textual Entailment (RTE) datasets come from a series of annual textual
    entailment challenges (RTE1, RTE2, RTE3 and RTE5).

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_rte"
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> rte_val = gluonnlp.data.SuperGlueRTE('val', root='./datasets/rte')
    -etc-
    >>> len(rte_val)
    277
    >>> sorted(rte_val[0].items())
    [('hypothesis', 'Christopher Reeve had an accident.'), ('idx', 0), ('label', 'not_entailment'), ('premise', 'Dana Reeve, the widow of the actor Christopher Reeve, has died of lung cancer at age 44, according to the Christopher Reeve Foundation.')]
    >>> rte_test = gluonnlp.data.SuperGlueRTE('test', root='./datasets/rte')
    -etc-
    >>> len(rte_test)
    3000
    >>> sorted(rte_test[0].items())
    [('hypothesis', 'Shukla is related to Mangla.'), ('idx', 0), ('premise', "Mangla was summoned after Madhumita's sister Nidhi Shukla, who was the first witness in the case.")]

    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_rte')):
        self._segment = segment
        self._data_file = {'train': ('train', 'a4471b47b23f6d8bc2e89b2ccdcf9a3a987c69a1',
                                     '01ebec38ff3d2fdd849d3b33c2a83154d1476690'),
                           'val': ('val', '17f23360f77f04d03aee6c42a27a61a6378f1fd9',
                                   '410f8607d9fc46572c03f5488387327b33589069'),
                           'test': ('test', 'ef2de5f8351ef80036c4aeff9f3b46106b4f2835',
                                    '69f9d9b4089d0db5f0605eeaebc1c7abc044336b')}
        data_file = self._data_file[segment]

        super(SuperGlueRTE, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/RTE'


@register(segment=['train', 'val', 'test'])
class SuperGlueCB(_SuperGlueDataset):
    """The CommitmentBank (CB) is a corpus of short texts in which at least one sentence
    contains an embedded clause.

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_cb"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> cb_val = gluonnlp.data.SuperGlueCB('val', root='./datasets/cb')
    -etc-
    >>> len(cb_val)
    56
    >>> sorted(cb_val[0].items())
    [('hypothesis', 'Valence was helping'), ('idx', 0), ('label', 'contradiction'), ('premise', "Valence the void-brain, Valence the virtuous valet. Why couldn't the figger choose his own portion of titanic anatomy to shaft? Did he think he was helping?")]
    >>> cb_test = gluonnlp.data.SuperGlueCB('test', root='./datasets/cb')
    -etc-
    >>> len(cb_test)
    250
    >>> sorted(cb_test[0].items())
    [('hypothesis', 'Polly was not an experienced ocean sailor'), ('idx', 0), ('premise', 'Polly had to think quickly. They were still close enough to shore for him to return her to the police if she admitted she was not an experienced ocean sailor.')]
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_cb')):
        self._segment = segment
        self._data_file = {'train': ('train', '0b27cbdbbcdf2ba82da2f760e3ab40ed694bd2b9',
                                     '193bdb772d2fe77244e5a56b4d7ac298879ec529'),
                           'val': ('val', 'e1f9dc77327eba953eb41d5f9b402127d6954ae0',
                                   'd286ac7c9f722c2b660e764ec3be11bc1e1895f8'),
                           'test': ('test', '008f9afdc868b38fdd9f989babe034a3ac35dd06',
                                    'cca70739162d54f3cd671829d009a1ab4fd8ec6a')}
        data_file = self._data_file[segment]

        super(SuperGlueCB, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/CB'


@register(segment=['train', 'val', 'test'])
class SuperGlueWSC(_SuperGlueDataset):
    """
    The Winograd Schema Challenge (WSC) is a co-reference resolution dataset.
    (Levesque et al., 2012)

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_wsc"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> wsc_val = gluonnlp.data.SuperGlueWSC('val', root='./datasets/wsc')
    -etc-
    >>> len(wsc_val)
    104
    >>> sorted(wsc_val[5].items())
    [('idx', 5), ('label', True), ('target', OrderedDict([('span2_index', 9), ('span1_index', 6), ('span1_text', 'The table'), ('span2_text', 'it')])), ('text', 'The large ball crashed right through the table because it was made of styrofoam.')]
    >>> wsc_test = gluonnlp.data.SuperGlueWSC('test', root='./datasets/wsc')
    -etc-
    >>> len(wsc_test)
    146
    >>> sorted(wsc_test[16].items())
    [('idx', 16), ('target', OrderedDict([('span1_text', 'life'), ('span1_index', 1), ('span2_text', 'it'), ('span2_index', 21)])), ('text', 'Your life is yours and yours alone, and if the pain outweighs the benefit, you should have the option to end it .')]
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_wsc')):
        self._segment = segment
        self._data_file = {'train': ('train', 'ed0fe96914cfe1ae8eb9978877273f6baed621cf',
                                     'fa978f6ad4b014b5f5282dee4b6fdfdaeeb0d0df'),
                           'val': ('val', 'cebec2f5f00baa686573ae901bb4d919ca5d3483',
                                   'ea2413e4e6f628f2bb011c44e1d8bae301375211'),
                           'test': ('test', '3313896f315e0cb2bb1f24f3baecec7fc93124de',
                                    'a47024aa81a5e7c9bc6e957b36c97f1d1b5da2fd')}
        data_file = self._data_file[segment]

        super(SuperGlueWSC, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/WSC'


@register(segment=['train', 'val', 'test'])
class SuperGlueWiC(_SuperGlueDataset):
    """
    The Word-in-Context (WiC) is a word sense disambiguation dataset cast as binary classification
    of sentence pairs. (Pilehvar and Camacho-Collados, 2019)

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_wic"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> wic_val = gluonnlp.data.SuperGlueWiC('val', root='./datasets/wic')
    -etc-
    >>> len(wic_val)
    638
    >>> sorted(wic_val[3].items())
    [('end1', 31), ('end2', 35), ('idx', 3), ('label', True), ('sentence1', 'She gave her hair a quick brush.'), ('sentence2', 'The dentist recommended two brushes a day.'), ('start1', 26), ('start2', 28), ('version', 1.1), ('word', 'brush')]
    >>> wic_test = gluonnlp.data.SuperGlueWiC('test', root='./datasets/wic')
    -etc-
    >>> len(wic_test)
    1400
    >>> sorted(wic_test[0].items())
    [('end1', 46), ('end2', 22), ('idx', 0), ('sentence1', 'The smell of fried onions makes my mouth water.'), ('sentence2', 'His eyes were watering.'), ('start1', 41), ('start2', 14), ('version', 1.1), ('word', 'water')]
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_wic')):
        self._segment = segment
        self._data_file = {'train': ('train', 'ec1e265bbdcde1d8da0b56948ed30d86874b1f12',
                                     '831a58c553def448e1b1d0a8a36e2b987c81bc9c'),
                           'val': ('val', '2046c43e614d98d538a03924335daae7881f77cf',
                                   '73b71136a2dc2eeb3be7ab455a08f20b8dbe7526'),
                           'test': ('test', '77af78a49aac602b7bbf080a03b644167b781ba9',
                                    '1be93932d46c8f8dc665eb7af6703c56ca1b1e08')}
        data_file = self._data_file[segment]

        super(SuperGlueWiC, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/WiC'


@register(segment=['train', 'val', 'test'])
class SuperGlueCOPA(_SuperGlueDataset):
    """
    The Choice of Plausible Alternatives (COPA) is a causal reasoning dataset.
    (Roemmele et al., 2011)

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_copa"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> copa_val = gluonnlp.data.SuperGlueCOPA('val', root='./datasets/copa')
    -etc-
    >>> len(copa_val)
    100
    >>> sorted(copa_val[0].items())
    [('choice1', 'The toilet filled with water.'), ('choice2', 'Water flowed from the spout.'), ('idx', 0), ('label', 1), ('premise', 'The man turned on the faucet.'), ('question', 'effect')]
    >>> copa_test = gluonnlp.data.SuperGlueCOPA('test', root='./datasets/copa')
    -etc-
    >>> len(copa_test)
    500
    >>> sorted(copa_test[0].items())
    [('choice1', 'It was fragile.'), ('choice2', 'It was small.'), ('idx', 0), ('premise', 'The item was packaged in bubble wrap.'), ('question', 'cause')]
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_copa')):
        self._segment = segment
        self._data_file = {'train': ('train', '96d20163fa8371e2676a50469d186643a07c4e7b',
                                     '5bb9c8df7b165e831613c8606a20cbe5c9622cc3'),
                           'val': ('val', 'acc13ad855a1d2750a3b746fb0cfe3ca6e8b6615',
                                   'c8b908d880ffaf69bd897d6f2a1f23b8c3a732d4'),
                           'test': ('test', '89347d7884e71b49dd73c6bcc317aef64bb1bac8',
                                    '735f39f3d31409d83b16e56ad8aed7725ef5ddd5')}
        data_file = self._data_file[segment]

        super(SuperGlueCOPA, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/COPA'


@register(segment=['train', 'val', 'test'])
class SuperGlueMultiRC(_SuperGlueDataset):
    """
    Multi-Sentence Reading Comprehension (MultiRC) is a QA dataset.
    (Khashabi et al., 2018)

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_multirc"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> multirc_val = gluonnlp.data.SuperGlueMultiRC('val', root='./datasets/multirc')
    -etc-
    >>> len(multirc_val)
    83
    >>> sorted(multirc_val[0].keys())
    ['questions', 'text']
    >>> len(multirc_val[0]['text'])
    12
    >>> len(multirc_val[0]['questions'])
    13
    >>> sorted(multirc_val[0]['questions'][0].keys())
    ['answers', 'idx', 'multisent', 'question', 'sentences_used']
    >>> multirc_test = gluonnlp.data.SuperGlueMultiRC('test', root='./datasets/multirc')
    -etc-
    >>> len(multirc_test)
    166
    >>> sorted(multirc_test[0].keys())
    ['questions', 'text']
    >>> len(multirc_test[0]['text'])
    14
    >>> len(multirc_test[0]['questions'])
    14
    >>> sorted(multirc_test[0]['questions'][0].keys())
    ['answers', 'idx', 'multisent', 'question', 'sentences_used']
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_multirc')):
        self._segment = segment
        self._data_file = {'train': ('train', '28d908566004fb84ff81828db8955f86fb771929',
                                     '2ef471a038f0b8116bf056da6440f290be7ab96e'),
                           'val': ('val', 'af93161bb987fbafe68111bce87dece4472b4ca0',
                                   '2364ed153f4f4e8cadde78680229a8544ba427db'),
                           'test': ('test', 'eabf1e8b426a8370cd3755a99412c7871a47ffa4',
                                    'd6d1107520d535332969ffe5f5b9bd7af2a33072')}
        data_file = self._data_file[segment]

        super(SuperGlueMultiRC, self).__init__(root, data_file)

    def _read_samples(self, samples):
        for i, sample in enumerate(samples):
            paragraph = dict()
            text = sample['paragraph']['text']
            sentences = self._split_text(text)
            paragraph['text'] = sentences
            paragraph['questions'] = sample['paragraph']['questions']
            samples[i] = paragraph
        return samples

    def _split_text(self, text):
        text = re.sub(r'<b>Sent .{1,2}: </b>', '', text)
        text = text.split('<br>')
        sents = [s for s in text if len(s) > 0]
        return sents

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/MultiRC'


@register(segment=['train', 'val', 'test'])
class SuperGlueBoolQ(_SuperGlueDataset):
    """
    Boolean Questions (BoolQ) is a QA dataset where each example consists of a short
    passage and a yes/no question about it.

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_boolq"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> boolq_val = gluonnlp.data.SuperGlueBoolQ('val', root='./datasets/boolq')
    -etc-
    >>> len(boolq_val)
    3270
    >>> sorted(boolq_val[0].keys())
    ['idx', 'label', 'passage', 'question']
    >>> boolq_test = gluonnlp.data.SuperGlueBoolQ('test', root='./datasets/boolq')
    -etc-
    >>> len(boolq_test)
    3245
    >>> sorted(boolq_test[0].keys())
    ['idx', 'passage', 'question']
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_boolq')):
        self._segment = segment
        self._data_file = {'train': ('train', '89507ff3015c3212b72318fb932cfb6d4e8417ef',
                                     'd5be523290f49fc0f21f4375900451fb803817c0'),
                           'val': ('val', 'fd39562fc2c9d0b2b8289d02a8cf82aa151d0ad4',
                                   '9b09ece2b1974e4da20f0173454ba82ff8ee1710'),
                           'test': ('test', 'a805d4bd03112366d548473a6848601c042667d3',
                                    '98c308620c6d6c0768ba093858c92e5a5550ce9b')}
        data_file = self._data_file[segment]

        super(SuperGlueBoolQ, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/BoolQ'


@register(segment=['train', 'val', 'test'])
class SuperGlueReCoRD(_SuperGlueDataset):
    """
    Reading Comprehension with Commonsense Reasoning Dataset (ReCoRD) is a multiple-choice
    QA dataset.

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    root : str, default "$MXNET_HOME/datasets/superglue_record"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> record_val = gluonnlp.data.SuperGlueReCoRD('val', root='./datasets/record')
    -etc-
    >>> len(record_val)
    7481
    >>> sorted(record_val[0].keys())
    ['idx', 'passage', 'qas', 'source']
    >>> record_test = gluonnlp.data.SuperGlueReCoRD('test', root='./datasets/record')
    -etc-
    >>> len(record_test)
    7484
    >>> sorted(record_test[0].keys())
    ['idx', 'passage', 'qas', 'source']
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'superglue_record')):
        self._segment = segment
        self._data_file = {'train': ('train', '047282c912535c9a3bcea519935fde882feb619d',
                                     '65592074cefde2ecd1b27ce7b35eb8beb86c691a'),
                           'val': ('val', '442d8470bff2c9295231cd10262a7abf401edc64',
                                   '9d1850e4dfe2eca3b71bfea191d5f4b412c65309'),
                           'test': ('test', 'fc639a18fa87befdc52f14c1092fb40475bf50d0',
                                    'b79b22f54b5a49f98fecd05751b122ccc6947c81')}
        data_file = self._data_file[segment]

        super(SuperGlueReCoRD, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/ReCoRD'


class SuperGlueAXb(_SuperGlueDataset):
    """
    The Broadcoverage Diagnostics (AX-b) is a diagnostics dataset labeled closely to
    the schema of MultiNLI.

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    root : str, default "$MXNET_HOME/datasets/superglue_ax_b"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> ax_b = gluonnlp.data.SuperGlueAXb(root='./datasets/ax_b')
    -etc-
    >>> len(ax_b)
    1104
    >>> sorted(ax_b[0].items())
    [('idx', '0'), ('label', 'not_entailment'), ('logic', 'Negation'), ('sentence1', 'The cat sat on the mat.'), ('sentence2', 'The cat did not sit on the mat.')]
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'superglue_ax_b')):
        data_file = ('AX-b', '398c5a376eb436f790723cd217ac040334140000',
                     '50fd8ac409897b652daa4b246917097c3c394bc8')

        super(SuperGlueAXb, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/AX-b'


class SuperGlueAXg(_SuperGlueDataset):
    """
    The Winogender Schema Diagnostics (AX-g) is a diagnostics dataset labeled closely to
    the schema of MultiNLI.

    From
    https://super.gluebenchmark.com/tasks

    Parameters
    ----------
    root : str, default "$MXNET_HOME/datasets/superglue_ax_g"
        Path to temp folder from storing data.
        MXNET_HOME defaults to '~/.mxnet'

    Examples
    --------
    >>> ax_g = gluonnlp.data.SuperGlueAXg(root='./datasets/ax_g')
    -etc-
    >>> len(ax_g)
    356
    >>> sorted(ax_g[0].items())
    [('hypothesis', 'The accountant sought help filing taxes.'), ('idx', 0), ('label', 'not_entailment'), ('pair_id', 551646), ('premise', 'The taxpayer met with the accountant to get help filing his taxes.')]
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'superglue_ax_g')):
        data_file = ('AX-g', 'd8c92498496854807dfeacd344eddf466d7f468a',
                     '8a8cbfe00fd88776a2a2f20b477e5b0c6cc8ebae')

        super(SuperGlueAXg, self).__init__(root, data_file)

    def _read_samples(self, samples):
        return samples

    def _repo_dir(self):
        return 'gluon/dataset/SUPERGLUE/AX-g'



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Weight updating functions."""
import os
import warnings
import numpy
from mxnet.optimizer import Optimizer, register
from mxnet.ndarray import zeros, NDArray, full
from mxnet.ndarray.contrib import mp_adamw_update, adamw_update, \
    multi_mp_adamw_update, multi_adamw_update

__all__ = ['BERTAdam']

@register
class BERTAdam(Optimizer):
    """The Adam optimizer with weight decay regularization for BERT.

    Updates are applied by::

        rescaled_grad = clip(grad * rescale_grad, clip_gradient)
        m = beta1 * m + (1 - beta1) * rescaled_grad
        v = beta2 * v + (1 - beta2) * (rescaled_grad**2)
        w = w - learning_rate * (m / (sqrt(v) + epsilon) + wd * w)

    Note that this is different from `mxnet.optimizer.Adam`, where L2 loss is added and
    accumulated in m and v. In BERTAdam, the weight decay term decoupled from gradient
    based update.

    This is also slightly different from the AdamW optimizer described in
    *Fixing Weight Decay Regularization in Adam*, where the schedule multiplier and
    learning rate is decoupled, and the bias-correction terms are removed.
    The BERTAdam optimizer uses the same learning rate to apply gradients
    w.r.t. the loss and weight decay.

    This optimizer accepts the following parameters in addition to those accepted
    by :class:`mxnet.optimizer.Optimizer`.

    Parameters
    ----------
    beta1 : float, optional, default is 0.9
        Exponential decay rate for the first moment estimates.
    beta2 : float, optional, default is 0.999
        Exponential decay rate for the second moment estimates.
    epsilon : float, optional, default is 1e-6
        Small value to avoid division by 0.
    """
    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, epsilon=1e-6,
                 **kwargs):
        super(BERTAdam, self).__init__(learning_rate=learning_rate, **kwargs)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.aggregate_num = max(1, min(50, int(os.getenv('MXNET_OPTIMIZER_AGGREGATION_SIZE',
                                                          '4'))))

    def create_state_multi_precision(self, index, weight):
        """multi-precision state creation function."""
        weight_master_copy = None
        if self.multi_precision and weight.dtype == numpy.float16:
            weight_master_copy = weight.astype(numpy.float32)
            return (self.create_state(index, weight_master_copy), weight_master_copy)
        if weight.dtype == numpy.float16 and not self.multi_precision:
            warnings.warn('Accumulating with float16 in optimizer can lead to '
                          'poor accuracy or slow convergence. '
                          'Consider using multi_precision=True option of the '
                          'BERTAdam optimizer')
        return self.create_state(index, weight)

    def create_state(self, _, weight):
        """state creation function."""
        return (zeros(weight.shape, weight.context, dtype=weight.dtype), #mean
                zeros(weight.shape, weight.context, dtype=weight.dtype)) #variance

    def update(self, index, weight, grad, state):
        """update function"""
        self._update_impl(index, weight, grad, state, multi_precision=False)

    def update_multi_precision(self, index, weight, grad, state):
        """multi-precision update function"""
        use_multi_precision = self.multi_precision and weight[0].dtype == numpy.float16
        self._update_impl(index, weight, grad, state,
                          multi_precision=use_multi_precision)

    def _update_impl(self, indices, weight, grad, state, multi_precision=False):
        """update function"""
        aggregate = self.aggregate_num > 1
        if not isinstance(indices, (tuple, list)):
            indices = [indices]
            weight = [weight]
            grad = [grad]
            state = [state]
        for w_i, g_i in zip(weight, grad):
            assert(isinstance(w_i, NDArray))
            assert(isinstance(g_i, NDArray))
            aggregate = (aggregate and
                         w_i.stype == 'default' and
                         g_i.stype == 'default')
        self._update_count(indices)
        lrs = self._get_lrs(indices)
        wds = self._get_wds(indices)

        # pylint: disable=access-member-before-definition
        if not isinstance(self.rescale_grad, NDArray):
            self.rescale_grad = full(shape=(1,), val=self.rescale_grad, ctx=weight[0].context)
        else:
            self.rescale_grad = self.rescale_grad.as_in_context(weight[0].context)

        kwargs = {'beta1': self.beta1, 'beta2': self.beta2, 'epsilon': self.epsilon,
                  'rescale_grad': self.rescale_grad}
        if self.clip_gradient:
            kwargs['clip_gradient'] = self.clip_gradient

        if aggregate:
            current_index = 0
            while current_index < len(indices):
                sidx = current_index
                eidx = min(current_index + self.aggregate_num, len(indices))
                if not multi_precision:
                    mean, var = list(zip(*state[sidx:eidx]))
                    multi_adamw_update(weight[sidx:eidx],
                                       grad[sidx:eidx],
                                       mean, var,
                                       out=weight[sidx:eidx],
                                       size=len(weight[sidx:eidx]),
                                       lrs=list(numpy.ones(len(weight[sidx:eidx]))),
                                       wds=wds[sidx:eidx],
                                       etas=lrs[sidx:eidx],
                                       **kwargs)
                else:
                    mean_var = list(zip(*state[sidx:eidx]))[0]
                    tmean_var = list(zip(*mean_var))
                    mean = tmean_var[0]
                    var = tmean_var[1]
                    multi_mp_adamw_update(weight[sidx:eidx],
                                          grad[sidx:eidx],
                                          mean, var,
                                          list(zip(*state[sidx:eidx]))[1],
                                          out=weight[sidx:eidx],
                                          size=len(weight[sidx:eidx]),
                                          lrs=list(numpy.ones(len(weight[sidx:eidx]))),
                                          wds=wds[sidx:eidx],
                                          etas=lrs[sidx:eidx],
                                          **kwargs)
                current_index += self.aggregate_num
        else:
            for w_i, g_i, s_i, lr, wd in zip(weight, grad, state, lrs, wds):
                if not multi_precision:
                    mean, var = s_i
                    adamw_update(w_i, g_i, mean, var, out=w_i,
                                 lr=1, wd=wd, eta=lr, **kwargs)
                else:
                    mean, var = s_i[0]
                    mp_adamw_update(w_i, g_i, mean, var, s_i[1], out=w_i,
                                    lr=1, wd=wd, eta=lr, **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""LSTM projection cell with cell clip and projection clip."""
__all__ = ['LSTMPCellWithClip']

from mxnet.gluon.contrib.rnn import LSTMPCell


class LSTMPCellWithClip(LSTMPCell):
    r"""Long-Short Term Memory Projected (LSTMP) network cell with cell clip and projection clip.
    Each call computes the following function:

    .. math::

        \DeclareMathOperator{\sigmoid}{sigmoid}
        \begin{array}{ll}
        i_t = \sigmoid(W_{ii} x_t + b_{ii} + W_{ri} r_{(t-1)} + b_{ri}) \\
        f_t = \sigmoid(W_{if} x_t + b_{if} + W_{rf} r_{(t-1)} + b_{rf}) \\
        g_t = \tanh(W_{ig} x_t + b_{ig} + W_{rc} r_{(t-1)} + b_{rg}) \\
        o_t = \sigmoid(W_{io} x_t + b_{io} + W_{ro} r_{(t-1)} + b_{ro}) \\
        c_t = c_{\text{clip}}(f_t * c_{(t-1)} + i_t * g_t) \\
        h_t = o_t * \tanh(c_t) \\
        r_t = p_{\text{clip}}(W_{hr} h_t)
        \end{array}

    where :math:`c_{\text{clip}}` is the cell clip applied on the next cell;
    :math:`r_t` is the projected recurrent activation at time `t`,
    :math:`p_{\text{clip}}` means apply projection clip on he projected output.
    math:`h_t` is the hidden state at time `t`, :math:`c_t` is the
    cell state at time `t`, :math:`x_t` is the input at time `t`, and :math:`i_t`,
    :math:`f_t`, :math:`g_t`, :math:`o_t` are the input, forget, cell, and
    out gates, respectively.

    Parameters
    ----------
    hidden_size : int
        Number of units in cell state symbol.
    projection_size : int
        Number of units in output symbol.
    i2h_weight_initializer : str or Initializer
        Initializer for the input weights matrix, used for the linear
        transformation of the inputs.
    h2h_weight_initializer : str or Initializer
        Initializer for the recurrent weights matrix, used for the linear
        transformation of the hidden state.
    h2r_weight_initializer : str or Initializer
        Initializer for the projection weights matrix, used for the linear
        transformation of the recurrent state.
    i2h_bias_initializer : str or Initializer, default 'lstmbias'
        Initializer for the bias vector. By default, bias for the forget
        gate is initialized to 1 while all other biases are initialized
        to zero.
    h2h_bias_initializer : str or Initializer
        Initializer for the bias vector.
    prefix : str
        Prefix for name of `Block`s
        (and name of weight if params is `None`).
    params : Parameter or None
        Container for weight sharing between cells.
        Created if `None`.
    cell_clip : float
        Clip cell state between `[-cell_clip, cell_clip]` in LSTMPCellWithClip cell
    projection_clip : float
        Clip projection between `[-projection_clip, projection_clip]` in LSTMPCellWithClip cell
    """
    def __init__(self, hidden_size, projection_size,
                 i2h_weight_initializer=None, h2h_weight_initializer=None,
                 h2r_weight_initializer=None,
                 i2h_bias_initializer='zeros', h2h_bias_initializer='zeros',
                 input_size=0, cell_clip=None, projection_clip=None, prefix=None, params=None):
        super(LSTMPCellWithClip, self).__init__(hidden_size,
                                                projection_size,
                                                i2h_weight_initializer,
                                                h2h_weight_initializer,
                                                h2r_weight_initializer,
                                                i2h_bias_initializer,
                                                h2h_bias_initializer,
                                                input_size,
                                                prefix=prefix,
                                                params=params)

        self._cell_clip = cell_clip
        self._projection_clip = projection_clip

    # pylint: disable= arguments-differ
    def hybrid_forward(self, F, inputs, states, i2h_weight,
                       h2h_weight, h2r_weight, i2h_bias, h2h_bias):
        r"""Hybrid forward computation for Long-Short Term Memory Projected network cell
        with cell clip and projection clip.

        Parameters
        ----------
        inputs : input tensor with shape `(batch_size, input_size)`.
        states : a list of two initial recurrent state tensors, with shape
            `(batch_size, projection_size)` and `(batch_size, hidden_size)` respectively.

        Returns
        --------
        out : output tensor with shape `(batch_size, num_hidden)`.
        next_states : a list of two output recurrent state tensors. Each has
            the same shape as `states`.
        """
        prefix = 't%d_'%self._counter
        i2h = F.FullyConnected(data=inputs, weight=i2h_weight, bias=i2h_bias,
                               num_hidden=self._hidden_size*4, name=prefix+'i2h')
        h2h = F.FullyConnected(data=states[0], weight=h2h_weight, bias=h2h_bias,
                               num_hidden=self._hidden_size*4, name=prefix+'h2h')
        gates = i2h + h2h
        slice_gates = F.SliceChannel(gates, num_outputs=4, name=prefix+'slice')
        in_gate = F.Activation(slice_gates[0], act_type='sigmoid', name=prefix+'i')
        forget_gate = F.Activation(slice_gates[1], act_type='sigmoid', name=prefix+'f')
        in_transform = F.Activation(slice_gates[2], act_type='tanh', name=prefix+'c')
        out_gate = F.Activation(slice_gates[3], act_type='sigmoid', name=prefix+'o')
        next_c = F._internal._plus(forget_gate * states[1], in_gate * in_transform,
                                   name=prefix+'state')
        if self._cell_clip is not None:
            next_c = next_c.clip(-self._cell_clip, self._cell_clip)
        hidden = F._internal._mul(out_gate, F.Activation(next_c, act_type='tanh'),
                                  name=prefix+'hidden')
        next_r = F.FullyConnected(data=hidden, num_hidden=self._projection_size,
                                  weight=h2r_weight, no_bias=True, name=prefix+'out')
        if self._projection_clip is not None:
            next_r = next_r.clip(-self._projection_clip, self._projection_clip)

        return next_r, [next_r, next_c]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Utility functions for trainer and parameters."""

__all__ = ['grad_global_norm', 'clip_grad_global_norm', 'save_parameters',
           'save_states', 'load_parameters', 'load_states']

import warnings

from collections import defaultdict
import mxnet as mx
from mxnet import nd
from .. import _constants as C
from .files import _TempFilePath, _transfer_file_s3

def grad_global_norm(parameters, max_norm=None):
    """Calculate the 2-norm of gradients of parameters, and how much they should be scaled down
    such that their 2-norm does not exceed `max_norm`, if `max_norm` if provided.

    If gradients exist for more than one context for a parameter, user needs to explicitly call
    ``trainer.allreduce_grads`` so that the gradients are summed first before calculating
    the 2-norm.

    .. note::
        This function is only for use when `update_on_kvstore` is set to False in trainer.

    Example::

        trainer = Trainer(net.collect_params(), update_on_kvstore=False, ...)
        for x, y in mx.gluon.utils.split_and_load(X, [mx.gpu(0), mx.gpu(1)]):
            with mx.autograd.record():
                y = net(x)
                loss = loss_fn(y, label)
            loss.backward()
        trainer.allreduce_grads()
        norm = grad_global_norm(net.collect_params().values())
        ...

    Parameters
    ----------
    parameters : list of Parameters
    max_norm: NDArray, optional
        The maximum L2 norm threshold. If provided, `ratio` and `is_finite` will be returned.

    Returns
    -------
    NDArray
      Total norm. Shape is (1,)
    NDArray
      Ratio for rescaling gradients based on max_norm s.t. grad = grad / ratio.
      If total norm is NaN, ratio will be NaN, too.
      Returned if `max_norm` is provided. Shape is (1,)
    NDArray
      Whether the total norm is finite, returned if `max_norm` is provided. Shape is (1,)
    """
    # distribute gradients among contexts
    idx = 0
    arrays = defaultdict(list)
    sum_norms = []
    for p in parameters:
        if p.grad_req != 'null':
            p_grads = p.list_grad()
            arrays[idx % len(p_grads)].append(p_grads[idx % len(p_grads)])
            idx += 1
    assert len(arrays) > 0, 'No parameter found available for gradient norm.'

    ctx, dtype = arrays[0][0].context, 'float32'
    for idx, arr in enumerate(arrays.values()):
        sum_norm = mx.nd.multi_sum_sq(*arr, num_arrays=len(arr))
        sum_norm = nd.add_n(*sum_norm)
        sum_norms.append(sum_norm.as_in_context(ctx))

    # reduce
    total_norm = nd.add_n(*sum_norms).sqrt()
    if max_norm is None:
        return total_norm
    scale = total_norm / max_norm
    # is_finite = 0 if NaN or Inf, 1 otherwise.
    is_finite = nd.contrib.isfinite(scale)
    # if scale is finite, nd.maximum selects the max between scale and 1. That is,
    # 1 is returned if total_norm does not exceed max_norm.
    # if scale = NaN or Inf, the result of nd.minimum is undefined. Therefore, we use
    # choices.take to return NaN or Inf.
    scale_or_one = nd.maximum(nd.ones((1,), dtype=dtype, ctx=ctx), scale)
    choices = nd.concat(scale, scale_or_one, dim=0)
    chosen_scale = choices.take(is_finite)
    return total_norm, chosen_scale, is_finite

def clip_grad_global_norm(parameters, max_norm, check_isfinite=True):
    """Rescales gradients of parameters so that the sum of their 2-norm is smaller than `max_norm`.
    If gradients exist for more than one context for a parameter, user needs to explicitly call
    ``trainer.allreduce_grads`` so that the gradients are summed first before calculating
    the 2-norm.

    .. note::

        This function is only for use when `update_on_kvstore` is set to False in trainer.
        In cases where training happens on multiple contexts, this method should be used in
        conjunction with ``trainer.allreduce_grads()`` and ``trainer.update()``.
        (**not** ``trainer.step()``)

    Example::

        trainer = Trainer(net.collect_params(), update_on_kvstore=False, ...)
        for x, y in mx.gluon.utils.split_and_load(X, [mx.gpu(0), mx.gpu(1)]):
            with mx.autograd.record():
                y = net(x)
                loss = loss_fn(y, label)
            loss.backward()
        trainer.allreduce_grads()
        nlp.utils.clip_grad_global_norm(net.collect_params().values(), max_norm)
        trainer.update(batch_size)
        ...

    Parameters
    ----------
    parameters : list of Parameters
    max_norm : float
    check_isfinite : bool, default True
         If True, check that the total_norm is finite (not nan or inf). This
         requires a blocking .asscalar() call.

    Returns
    -------
    NDArray or float
      Total norm. Return type is NDArray of shape (1,) if check_isfinite is
      False. Otherwise a float is returned.

    """
    total_norm, ratio, is_finite = grad_global_norm(parameters, max_norm)
    scale = 1 / ratio
    if check_isfinite:
        if is_finite != 1:
            warnings.warn(
                UserWarning('nan or inf is detected. '
                            'Clipping results will be undefined.'), stacklevel=2)
    for p in parameters:
        if p.grad_req != 'null':
            for arr in p.list_grad():
                arr *= scale.as_in_context(arr.context)
    return total_norm

def _s3_compatible_save_load(is_save, save_load_method, filename, *args, **kwargs):
    """Dispatch function for save load with s3."""
    if C.S3_PREFIX in filename:
        # create temp dir
        with _TempFilePath() as temp_path:
            if is_save:
                # save model
                save_load_method(temp_path, *args, **kwargs)
                _transfer_file_s3(temp_path, filename, upload=is_save)
            else:
                # load model
                _transfer_file_s3(temp_path, filename, upload=is_save)
                save_load_method(temp_path, *args, **kwargs)
    else:
        save_load_method(filename, *args, **kwargs)

def load_parameters(model, filename, ctx=None, allow_missing=False,
                    ignore_extra=False, cast_dtype=None):
    """Load parameters from file previously saved by `save_parameters`.

    Both local file system path and S3 URI are supported.
    For example, 's3://mybucket/folder/net.params', './folder/net.params'.

    Parameters
    ----------
    filename : str
        Path to parameter file.
    ctx : Context or list of Context, default cpu()
        Context(s) to initialize loaded parameters on.
    allow_missing : bool, default False
        Whether to silently skip loading parameters not represents in the file.
    ignore_extra : bool, default False
        Whether to silently ignore parameters from the file that are not
        present in this Block.
    cast_dtype : bool, default False
        Cast the data type of the NDArray loaded from the checkpoint to the dtype
        provided by the Parameter if any.
    """
    if cast_dtype is not None:
        _s3_compatible_save_load(False, model.load_parameters, filename, ctx=ctx,
                                 allow_missing=allow_missing, ignore_extra=ignore_extra,
                                 cast_dtype=cast_dtype)
    else:
        _s3_compatible_save_load(False, model.load_parameters, filename, ctx=ctx,
                                 allow_missing=allow_missing, ignore_extra=ignore_extra)

def save_parameters(model, filename):
    """Save parameters to file.

    Saved parameters can only be loaded with `Block.load_parameters`. Note that this
    method only saves parameters, not model structure.

    Both local file system path and S3 URI are supported.
    For example, 's3://mybucket/folder/net.params', './folder/net.params'.

    Parameters
    ----------
    model : mx.gluon.Block
        The model to save.
    uri : str
        Path to file.
    """
    _s3_compatible_save_load(True, model.save_parameters, filename)


def load_states(trainer, fname):
    """Loads trainer states (e.g. optimizer, momentum) from a file.

    Both local file system path and S3 URI are supported.
    For example, 's3://mybucket/folder/net.states', './folder/net.states'.

    Parameters
    ----------
    trainer : mxnet.gluon.Trainer
        The trainer whose states will be loaded.
    fname : str
        Path to input states file.

    Note
    ----
    `optimizer.param_dict`, which contains Parameter information (such as
    `lr_mult` and `wd_mult`) will not be loaded from the file, but rather set
    based on current Trainer's parameters.
    """
    _s3_compatible_save_load(False, trainer.load_states, fname)

def save_states(trainer, fname):
    """Saves trainer states (e.g. optimizer, momentum) to a file.

    Both local file system path and S3 URI are supported.
    For example, 's3://mybucket/folder/net.states', './folder/net.states'.

    Parameters
    ----------
    trainer : mxnet.gluon.Trainer
        The trainer whose states will be saved.
    fname : str
        Path to output states file.

    Note
    ----
    `optimizer.param_dict`, which contains Parameter information (such as
    `lr_mult` and `wd_mult`) will not be saved.
    """
    _s3_compatible_save_load(True, trainer.save_states, fname)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Label Smoothing"""

__all__ = ['LabelSmoothing']

import mxnet as mx
from mxnet.gluon import HybridBlock

# pylint: disable=unused-argument
class _SmoothingWithDim(mx.operator.CustomOp):
    def __init__(self, epsilon=0.1, axis=-1):
        super(_SmoothingWithDim, self).__init__(True)
        self._epsilon = epsilon
        self._axis = axis

    def forward(self, is_train, req, in_data, out_data, aux):
        inputs = in_data[0]
        outputs = ((1 - self._epsilon) * inputs) + (self._epsilon / float(inputs.shape[self._axis]))
        self.assign(out_data[0], req[0], outputs)

    def backward(self, req, out_grad, in_data, out_data, in_grad, aux):
        self.assign(in_grad[0], req[0], (1 - self._epsilon) * out_grad[0])


@mx.operator.register('_smoothing_with_dim')
class _SmoothingWithDimProp(mx.operator.CustomOpProp):
    def __init__(self, epsilon=0.1, axis=-1):
        super(_SmoothingWithDimProp, self).__init__(True)
        self._epsilon = float(epsilon)
        self._axis = int(axis)

    def list_arguments(self):
        return ['data']

    def list_outputs(self):
        return ['output']

    def infer_shape(self, in_shape):
        data_shape = in_shape[0]
        output_shape = data_shape
        return (data_shape,), (output_shape,), ()

    def declare_backward_dependency(self, out_grad, in_data, out_data):
        return out_grad

    def create_operator(self, ctx, in_shapes, in_dtypes):
        #  create and return the CustomOp class.
        return _SmoothingWithDim(self._epsilon, self._axis)
# pylint: enable=unused-argument


class LabelSmoothing(HybridBlock):
    """Applies label smoothing. See https://arxiv.org/abs/1512.00567.

    It changes the construction of the probability to (1 - epsilon) for the true class,
    epsilon / (num_classes - 1) otherwise.

    Parameters
    ----------
    axis : int, default -1
        The axis to smooth.
    epsilon : float, default 0.1
        The epsilon parameter in label smoothing
    sparse_label : bool, default True
        Whether input is an integer array instead of one hot array.
    units : int or None
        Vocabulary size. If units is not given, it will be inferred from the input.
    prefix : str
        Prefix for name of `Block`s
        (and name of weight if params is `None`).
    params : Parameter or None
        Container for weight sharing between cells.
        Created if `None`.
    """
    def __init__(self, axis=-1, epsilon=0.1, units=None,
                 sparse_label=True, prefix=None, params=None):
        super(LabelSmoothing, self).__init__(prefix=prefix, params=params)
        self._axis = axis
        self._epsilon = epsilon
        self._sparse_label = sparse_label
        self._units = units

    def hybrid_forward(self, F, inputs, units=None): # pylint: disable=arguments-differ
        """

        Parameters
        ----------
        inputs : Symbol or NDArray
            Shape (batch_size, length) or (batch_size, length, V)
        units : int or None

        Returns
        -------
        smoothed_label : Symbol or NDArray
            Shape (batch_size, length, V)
        """
        if self._sparse_label:
            assert units is not None or self._units is not None, \
                'units needs to be given in function call or ' \
                'instance initialization when sparse_label is False'
            if units is None:
                units = self._units
            inputs = F.one_hot(inputs, depth=units)
        if units is None and self._units is None:
            return F.Custom(inputs, epsilon=self._epsilon, axis=self._axis,
                            op_type='_smoothing_with_dim')
        else:
            if units is None:
                units = self._units
            return ((1 - self._epsilon) * inputs) + (self._epsilon / units)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Datasets for intent classification and slot labeling."""

import io
import os
import zipfile
import numpy as np
from mxnet.gluon.utils import download, check_sha1, _get_repo_file_url
from mxnet.gluon.data import SimpleDataset

from .dataset import TSVDataset
from .registry import register
from .utils import Splitter
from ..base import get_home_dir
from ..vocab import Vocab

__all__ = ['ATISDataset', 'SNIPSDataset']


class _BaseICSLDataset(SimpleDataset):
    """Base Class of Datasets for Joint Intent Classification and Slot Labeling.

    Parameters
    ----------
    segment : str
        Dataset segment.
    root : str
        Path to temp folder for storing data.
    """
    def __init__(self, segment, root):
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._segment = segment
        self._root = root
        self._intent_vocab = None
        self._slot_vocab = None
        self._get_data()
        super(_BaseICSLDataset, self).__init__(self._read_data(segment))

    @property
    def _download_info(self):
        """Download file information.

        Returns
        -------
        filename_format : str
            The filename format with slot for short hash.
        sha1_hash : str
            Expected sha1 hash of the file content.
        """
        raise NotImplementedError

    @property
    def intent_vocab(self):
        if self._intent_vocab is None:
            with open(os.path.join(self._root, 'intent_vocab.json'), 'r') as f:
                self._intent_vocab = Vocab.from_json(f.read())
        return self._intent_vocab

    @property
    def slot_vocab(self):
        if self._slot_vocab is None:
            with open(os.path.join(self._root, 'slot_vocab.json'), 'r') as f:
                self._slot_vocab = Vocab.from_json(f.read())
        return self._slot_vocab

    def _get_data(self):
        filename_format, sha1_hash = self._download_info
        filename = filename_format.format(sha1_hash[:8])
        data_filename = os.path.join(self._root, filename)
        url = _get_repo_file_url('gluon/dataset', filename)
        if not os.path.exists(data_filename) or not check_sha1(data_filename, sha1_hash):
            download(url, path=data_filename, sha1_hash=sha1_hash)
            with zipfile.ZipFile(data_filename, 'r') as zf:
                zf.extractall(self._root)

    def _read_data(self, segment):
        sentences = TSVDataset(os.path.join(self._root, '{}_sentence.txt'.format(segment)),
                               field_separator=Splitter(' '))
        tags = TSVDataset(os.path.join(self._root, '{}_tags.txt'.format(segment)),
                          field_separator=Splitter(' '))
        with io.open(os.path.join(self._root, '{}_intent.txt'.format(segment)), 'r',
                     encoding='utf-8') as f:
            intents = []
            for line in f:
                line = line.strip()
                intents.append(np.array([self.intent_vocab[ele] for ele in line.split(';')],
                                        dtype=np.int32))
        return list(zip(sentences, tags, intents))


@register(segment=['train', 'dev', 'test'])
class ATISDataset(_BaseICSLDataset):
    """Airline Travel Information System dataset from MS CNTK.

    From
    https://github.com/Microsoft/CNTK/tree/master/Examples/LanguageUnderstanding/ATIS/Data

    License: Unspecified

    Each sample has three fields: tokens, slot labels, intent label.

    Parameters
    ----------
    segment : {'train', 'dev', 'test'}, default 'train'
        Dataset segment.
    root : str, default '$MXNET_HOME/datasets/atis'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> atis = gluonnlp.data.ATISDataset(root='./datasets/atis')
    -etc-
    >>> len(atis)
    4478
    >>> len(atis[0])
    3
    >>> len(atis[0][0])
    10
    >>> atis[0][0]
    ['i', 'want', 'to', 'fly', 'from', 'baltimore', 'to', 'dallas', 'round', 'trip']
    >>> len(atis[0][1])
    10
    >>> atis[0][1][:8]
    ['O', 'O', 'O', 'O', 'O', 'B-fromloc.city_name', 'O', 'B-toloc.city_name']
    >>> atis[0][2]
    array([10], dtype=int32)
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'atis')):
        super(ATISDataset, self).__init__(segment, root)

    @property
    def _download_info(self):
        return 'atis-{}.zip', 'fb75a1b595566d5c5ec06ee6f2296d6629b8c225'


@register(segment=['train', 'dev', 'test'])
class SNIPSDataset(_BaseICSLDataset):
    """Snips Natural Language Understanding Benchmark dataset.

    Coucke et al. (2018). Snips Voice Platform: an embedded Spoken Language Understanding system
    for private-by-design voice interfaces. https://arxiv.org/abs/1805.10190

    From
    https://github.com/snipsco/nlu-benchmark/tree/master/2017-06-custom-intent-engines

    License: Unspecified

    Each sample has three fields: tokens, slot labels, intent label.

    Parameters
    ----------
    segment : {'train', 'dev', 'test'}, default 'train'
        Dataset segment.
    root : str, default '$MXNET_HOME/datasets/snips'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> snips = gluonnlp.data.SNIPSDataset(root='./datasets/snips')
    -etc-
    >>> len(snips)
    13084
    >>> len(snips[0])
    3
    >>> len(snips[1][0])
    8
    >>> snips[1][0]
    ['put', 'United', 'Abominations', 'onto', 'my', 'rare', 'groove', 'playlist']
    >>> len(snips[1][1])
    8
    >>> snips[1][1][:5]
    ['O', 'B-entity_name', 'I-entity_name', 'O', 'B-playlist_owner']
    >>> snips[1][2]
    array([0], dtype=int32)
    """
    def __init__(self, segment='train',
                 root=os.path.join(get_home_dir(), 'datasets', 'snips')):
        super(SNIPSDataset, self).__init__(segment, root)

    @property
    def _download_info(self):
        return 'snips-{}.zip', 'f22420cc0f2a26078337dc375606be46a4cc8c51'



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=
"""Google billion words dataset."""

__all__ = ['GBWStream']

import glob
import hashlib
import io
import os
import tarfile
import zipfile

from mxnet.gluon.utils import _get_repo_file_url, check_sha1, download

from ..._constants import EOS_TOKEN
from ...base import get_home_dir
from ...vocab import Vocab
from ..dataset import CorpusDataset
from ..stream import SimpleDatasetStream


class GBWStream(SimpleDatasetStream):
    """1-Billion-Word word-level dataset for language modeling, from Google.

    The GBWSream iterates over CorpusDatasets(flatten=False).

    Source http://www.statmt.org/lm-benchmark

    License: Apache

    Parameters
    ----------
    segment : {'train', 'test'}, default 'train'
        Dataset segment.
    skip_empty : bool, default True
        Whether to skip the empty samples produced from sample_splitters. If False, `bos` and `eos`
        will be added in empty samples.
    bos : str or None, default None
        The token to add at the begining of each sentence. If None, nothing is added.
    eos : str or None, default '<eos>'
        The token to add at the end of each sentence. If None, nothing is added.
    root : str, default '$MXNET_HOME/datasets/gbw'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """

    _archive_data = ('1-billion-word-language-modeling-benchmark-r13output.tar.gz',
                     '4df859766482e12264a5a9d9fb7f0e276020447d')
    _archive_vocab = ('gbw-ebb1a287.zip',
                      '63b335dcc27b6804d0a14acb88332d2602fe0f59')
    _data_file = {'train': ('training-monolingual.tokenized.shuffled',
                            'news.en-00*-of-00100',
                            '5e0d7050b37a99fd50ce7e07dc52468b2a9cd9e8'),
                  'test': ('heldout-monolingual.tokenized.shuffled',
                           'news.en.heldout-00000-of-00050',
                           '0a8e2b7496ba0b5c05158f282b9b351356875445')}
    _vocab_file = ('gbw-ebb1a287.vocab',
                   'ebb1a287ca14d8fa6f167c3a779e5e7ed63ac69f')

    # Directory layout:
    # - root ($MXNET_HOME/datasets/gbw)
    #   - archive_file (1-billion-word-language-modeling-benchmark-r13output.tar.gz)
    #   - dir (1-billion-word-language-modeling-benchmark-r13output)
    #     - subdir (training-monolingual.tokenized.shuffled)
    #     - subdir (heldout-monolingual.tokenized.shuffled)

    def __init__(self, segment='train', skip_empty=True, bos=None, eos=EOS_TOKEN,
                 root=os.path.join(get_home_dir(), 'datasets', 'gbw')):
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._root = root
        self._dir = os.path.join(root, '1-billion-word-language-modeling-benchmark-r13output')
        self._namespace = 'gluon/dataset/gbw'
        subdir_name, pattern, data_hash = self._data_file[segment]
        self._subdir = os.path.join(self._dir, subdir_name)
        self._file_pattern = os.path.join(self._subdir, pattern)
        self._data_hash = data_hash
        self._get_data()
        sampler = 'sequential' if segment != 'train' else 'random'
        super().__init__(dataset=CorpusDataset, file_pattern=self._file_pattern,
                         skip_empty=skip_empty, bos=bos, eos=eos, file_sampler=sampler)

    def _get_data(self):
        archive_file_name, archive_hash = self._archive_data
        archive_file_path = os.path.join(self._root, archive_file_name)
        exists = False
        if os.path.exists(self._dir) and os.path.exists(self._subdir):
            # verify sha1 for all files in the subdir
            sha1 = hashlib.sha1()
            filenames = sorted(glob.glob(self._file_pattern))
            for filename in filenames:
                with open(filename, 'rb') as f:
                    while True:
                        data = f.read(1048576)
                        if not data:
                            break
                        sha1.update(data)
            if sha1.hexdigest() == self._data_hash:
                exists = True
        if not exists:
            # download archive
            if not os.path.exists(archive_file_path) or \
               not check_sha1(archive_file_path, archive_hash):
                download(_get_repo_file_url(self._namespace, archive_file_name),
                         path=self._root, sha1_hash=archive_hash)
            # extract archive
            with tarfile.open(archive_file_path, 'r:gz') as tf:
                tf.extractall(path=self._root)

    def _get_vocab(self):
        archive_file_name, archive_hash = self._archive_vocab
        vocab_file_name, vocab_hash = self._vocab_file
        namespace = 'gluon/dataset/vocab'
        root = self._root
        path = os.path.join(root, vocab_file_name)
        if not os.path.exists(path) or not check_sha1(path, vocab_hash):
            downloaded_path = download(_get_repo_file_url(namespace, archive_file_name),
                                       path=root, sha1_hash=archive_hash)

            with zipfile.ZipFile(downloaded_path, 'r') as zf:
                zf.extractall(path=root)
        return path

    @property
    def vocab(self):
        path = self._get_vocab()
        with io.open(path, 'r', encoding='utf-8') as in_file:
            return Vocab.from_json(in_file.read())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Building blocks and utility for models."""
__all__ = ['WeightDropParameter']

from mxnet import nd, gluon


class WeightDropParameter(gluon.Parameter):
    """A Container holding parameters (weights) of Blocks and performs dropout.

    Parameters
    ----------
    parameter : Parameter
        The parameter which drops out.
    rate : float, default 0.0
        Fraction of the input units to drop. Must be a number between 0 and 1.
        Dropout is not applied if dropout_rate is 0.
    mode : str, default 'training'
        Whether to only turn on dropout during training or to also turn on for inference.
        Options are 'training' and 'always'.
    axes : tuple of int, default ()
        Axes on which dropout mask is shared.
    """
    def __init__(self, parameter, rate=0.0, mode='training', axes=()):
        p = parameter
        self._deferred_init = p._deferred_init
        super(WeightDropParameter, self).__init__(
            name=p.name, grad_req=p.grad_req, shape=p._shape, dtype=p.dtype,
            lr_mult=p.lr_mult, wd_mult=p.wd_mult, init=p.init,
            allow_deferred_init=p._allow_deferred_init,
            differentiable=p._differentiable)
        self._rate = rate
        self._mode = mode
        self._axes = axes
        self._var = p._var
        self._data = p._data
        self._grad = p._grad
        self._ctx_list = p._ctx_list
        self._ctx_map = p._ctx_map
        self._trainer = p._trainer

    def data(self, ctx=None):
        """Returns a copy of this parameter on one context. Must have been
        initialized on this context before.

        Parameters
        ----------
        ctx : Context
            Desired context.
        Returns
        -------
        NDArray on ctx
        """
        d = self._check_and_get(self._data, ctx)
        if self._rate:
            d = nd.Dropout(d, self._rate, self._mode, self._axes)
        return d

    def __repr__(self):
        s = 'WeightDropParameter {name} (shape={shape}, dtype={dtype}, rate={rate}, mode={mode})'
        return s.format(name=self.name, shape=self.shape, dtype=self.dtype,
                        rate=self._rate, mode=self._mode)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=invalid-encoded-data, too-many-lines
"""Transformer API. It provides tools for common transformation on samples in text dataset, such as
clipping, padding, and tokenization."""


__all__ = [
    'ClipSequence', 'PadSequence', 'SacreMosesTokenizer',
    'SpacyTokenizer', 'SacreMosesDetokenizer',
    'JiebaTokenizer', 'NLTKStanfordSegmenter', 'SentencepieceTokenizer',
    'SentencepieceDetokenizer', 'BERTBasicTokenizer', 'BERTTokenizer',
    'BERTSentenceTransform', 'BERTSPTokenizer',
    'GPT2BPETokenizer', 'GPT2BPEDetokenizer'
]

import errno
import functools
import io
import os
import time
import unicodedata
import warnings
import zipfile
from typing import List, Optional

import mxnet as mx
from mxnet.gluon.utils import _get_repo_url, check_sha1, download
import numpy as np

from ..base import get_home_dir
from ..vocab.vocab import Vocab
from .utils import _extract_archive
from .fast_bert_tokenizer import is_control, is_punctuation, is_whitespace
from .fast_bert_tokenizer import BasicTokenizer, WordpieceTokenizer


class ClipSequence:
    """Clip the sequence to have length no more than `length`.

    Parameters
    ----------
    length : int
        Maximum length of the sequence

    Examples
    --------
    >>> datasets = gluon.data.SimpleDataset([[1, 3, 5, 7], [1, 2, 3], [1, 2, 3, 4, 5, 6, 7, 8]])
    >>> list(datasets.transform(gluonnlp.data.ClipSequence(4)))
    [[1, 3, 5, 7], [1, 2, 3], [1, 2, 3, 4]]
    >>> datasets = gluon.data.SimpleDataset([np.array([[1, 3], [5, 7], [7, 5], [3, 1]]),
    ...                                      np.array([[1, 2], [3, 4], [5, 6],
    ...                                                [6, 5], [4, 3], [2, 1]]),
    ...                                      np.array([[2, 4], [4, 2]])])
    >>> list(datasets.transform(gluonnlp.data.ClipSequence(3)))
    [array([[1, 3],
           [5, 7],
           [7, 5]]), array([[1, 2],
           [3, 4],
           [5, 6]]), array([[2, 4],
           [4, 2]])]
    """

    def __init__(self, length):
        self._length = length

    def __call__(self, sample):
        return sample[:min(len(sample), self._length)]


class PadSequence:
    """Pad the sequence.

    Pad the sequence to the given `length` by inserting `pad_val`. If `clip` is set,
    sequence that has length larger than `length` will be clipped.

    Parameters
    ----------
    length : int
        The maximum length to pad/clip the sequence
    pad_val : number
        The pad value. Default 0
    clip : bool

    Examples
    --------
    >>> datasets = gluon.data.SimpleDataset([[1, 3, 5, 7], [1, 2, 3], [1, 2, 3, 4, 5, 6, 7, 8]])
    >>> list(datasets.transform(gluonnlp.data.PadSequence(6)))
    [[1, 3, 5, 7, 0, 0], [1, 2, 3, 0, 0, 0], [1, 2, 3, 4, 5, 6]]
    >>> list(datasets.transform(gluonnlp.data.PadSequence(6, clip=False)))
    [[1, 3, 5, 7, 0, 0], [1, 2, 3, 0, 0, 0], [1, 2, 3, 4, 5, 6, 7, 8]]
    >>> list(datasets.transform(gluonnlp.data.PadSequence(6, pad_val=-1, clip=False)))
    [[1, 3, 5, 7, -1, -1], [1, 2, 3, -1, -1, -1], [1, 2, 3, 4, 5, 6, 7, 8]]
    """

    def __init__(self, length, pad_val=0, clip=True):
        self._length = length
        self._pad_val = pad_val
        self._clip = clip

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample : list of number or mx.nd.NDArray or np.ndarray

        Returns
        -------
        ret : list of number or mx.nd.NDArray or np.ndarray
        """
        sample_length = len(sample)
        if sample_length >= self._length:
            if self._clip and sample_length > self._length:
                return sample[:self._length]
            else:
                return sample
        else:
            if isinstance(sample, mx.nd.NDArray):
                # TODO(sxjscience) Use this trick for padding because mx.pad currently only supports
                # 4D/5D inputs
                new_sample_shape = (self._length, ) + sample.shape[1:]
                ret = mx.nd.full(shape=new_sample_shape,
                                 val=self._pad_val,
                                 ctx=sample.context,
                                 dtype=sample.dtype)
                ret[:sample_length] = sample
                return ret
            elif isinstance(sample, np.ndarray):
                pad_width = [(0, self._length - sample_length)] +\
                            [(0, 0) for _ in range(sample.ndim - 1)]
                return np.pad(sample,
                              mode='constant',
                              constant_values=self._pad_val,
                              pad_width=pad_width)
            elif isinstance(sample, list):
                return sample + [
                    self._pad_val for _ in range(self._length - sample_length)
                ]
            else:
                raise NotImplementedError(
                    'The input must be 1) list or 2) numpy.ndarray or 3) '
                    'mxnet.NDArray, received type=%s' % str(type(sample)))


class SacreMosesTokenizer:
    """Apply the Moses Tokenizer implemented in sacremoses.

    Users of this class are required to install
    `sacremoses <https://github.com/alvations/sacremoses>`_.
    For example, one can use :samp:`pip install sacremoses`.

    .. note::
        sacremoses carries an LGPL 2.1+ license.

    Examples
    --------
    >>> tokenizer = gluonnlp.data.SacreMosesTokenizer()
    >>> tokenizer('Gluon NLP toolkit provides a suite of text processing tools.')
    ['Gluon', 'NLP', 'toolkit', 'provides', 'a', 'suite', 'of', 'text', 'processing', 'tools', '.']
    >>> tokenizer('Das Gluon NLP-Toolkit stellt eine Reihe von Textverarbeitungstools '
    ...           'zur Verfügung.')
    ['Das', 'Gluon', 'NLP-Toolkit', 'stellt', 'eine', 'Reihe', 'von', 'Textverarbeitungstools', \
'zur', 'Verf\xfcgung', '.']
    """

    def __init__(self):
        from sacremoses import MosesTokenizer  # pylint: disable=import-outside-toplevel
        self._tokenizer = MosesTokenizer()

    def __call__(self, sample: str, return_str: bool = False):
        """Tokenize a sample.

        Parameters
        ----------
        sample
            The sentence to tokenize
        return_str
            True: return a single string
            False: return a list of tokens

        Returns
        -------
        ret : list of strs or str
            List of tokens or tokenized text
        """
        return self._tokenizer.tokenize(sample, return_str=return_str)


class SpacyTokenizer:
    """Apply the Spacy Tokenizer.

    Users of this class are required to install `spaCy <https://spacy.io/usage/>`_
    and download corresponding NLP models, such as :samp:`python -m spacy download en`.

    Only spacy>=2.0.0 is supported.

    Parameters
    ----------
    lang : str
        The language to tokenize. Default is 'en', i.e, English.
        You may refer to https://spacy.io/usage/models for supported languages.

    Examples
    --------
    >>> tokenizer = gluonnlp.data.SpacyTokenizer()
    >>> tokenizer('Gluon NLP toolkit provides a suite of text processing tools.')
    ['Gluon', 'NLP', 'toolkit', 'provides', 'a', 'suite', 'of', 'text', 'processing', 'tools', '.']
    >>> tokenizer = gluonnlp.data.SpacyTokenizer('de')
    >>> tokenizer('Das Gluon NLP-Toolkit stellt eine Reihe von Textverarbeitungstools'
    ...           ' zur Verfügung.')
    ['Das', 'Gluon', 'NLP-Toolkit', 'stellt', 'eine', 'Reihe', 'von', 'Textverarbeitungstools', \
'zur', 'Verf\xfcgung', '.']
    """

    def __init__(self, lang='en_core_web_sm'):
        try:
            import spacy  # pylint: disable=import-outside-toplevel
            from pkg_resources import parse_version  # pylint: disable=import-outside-toplevel
            assert parse_version(spacy.__version__) >= parse_version('2.0.0'),\
                'We only support spacy>=2.0.0'
        except ImportError:
            raise ImportError(
                'spaCy is not installed. You must install spaCy in order to use the '
                'SpacyTokenizer. You can refer to the official installation guide '
                'in https://spacy.io/usage/.')
        try:
            self._nlp = spacy.load(lang, disable=['parser', 'tagger', 'ner'])
        except IOError:
            raise IOError(
                'SpaCy Model for the specified language="{lang}" has not been '
                'downloaded. You need to check the installation guide in '
                'https://spacy.io/usage/models. Usually, the installation command '
                'should be `python -m spacy download {lang}`.'.format(
                    lang=lang))

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The sentence to tokenize

        Returns
        -------
        ret : list of strs
            List of tokens
        """
        return [tok.text for tok in self._nlp(sample)]


class SacreMosesDetokenizer:
    r"""Apply the Moses Detokenizer implemented in sacremoses.

    Users of this class are required to `install sacremoses
    <https://github.com/alvations/sacremoses>`_. For example, one can use
    :samp:`pip install sacremoses`.

    .. note::
        sacremoses carries an LGPL 2.1+ license.

    Parameters
    ----------
    return_str: bool, default False
        True: return a single string
        False: return a list of words

    Examples
    --------
    >>> detokenizer = gluonnlp.data.SacreMosesDetokenizer()
    >>> detokenizer(['Gluon', 'NLP', 'toolkit', 'provides', 'a', 'suite', 'of',
    ...              'text', 'processing', 'tools', '.'], return_str=True)
    'Gluon NLP toolkit provides a suite of text processing tools.'
    >>> detokenizer(['Das', 'Gluon','NLP-Toolkit','stellt','eine','Reihe','von',
    ...              'Textverarbeitungstools','zur','Verfügung','.'], return_str=True)
    'Das Gluon NLP-Toolkit stellt eine Reihe von Textverarbeitungstools zur Verfügung.'
    """

    def __init__(self, return_str=True):
        self._return_str = return_str
        from sacremoses import MosesDetokenizer  # pylint: disable=import-outside-toplevel
        self._detokenizer = MosesDetokenizer()

    def __call__(self, sample: List[str], return_str: Optional[bool] = None):
        """

        Parameters
        ----------
        sample
            The sentence to detokenize
        return_str
            True: return a single string
            False: return a list of words
            None: use constructor setting

        Returns
        -------
        ret : list of strs or str
            List of words or detokenized text
        """
        ret_str = self._return_str if return_str is None else return_str
        return self._detokenizer.detokenize(sample, return_str=ret_str)


class JiebaTokenizer:
    r"""Apply the jieba Tokenizer.

    Users of this class are required to `install jieba <https://github.com/fxsjy/jieba>`_

    Parameters
    ----------
    lang : str
        The language to tokenize. Default is "zh", i.e, Chinese.

    Examples
    --------
    >>> tokenizer = gluonnlp.data.JiebaTokenizer()
    >>> tokenizer('我来到北京清华大学')
    ['我', '来到', '北京', '清华大学']
    >>> tokenizer('小明硕士毕业于中国科学院计算所，后在日本京都大学深造')
    ['小明', '硕士', '毕业', '于', '中国科学院', '计算所', '，', '后', '在', '日本京都大学', '深造']

    """

    def __init__(self):
        try:
            with warnings.catch_warnings():  # jieba uses deprecated imp module
                warnings.simplefilter('ignore')
                import jieba  # pylint: disable=import-outside-toplevel
        except ImportError:
            raise ImportError(
                'jieba is not installed. You must install jieba in order to use the '
                'JiebaTokenizer. You can refer to the official installation guide '
                'in https://github.com/fxsjy/jieba')
        self._tokenizer = jieba

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The Chinese sentence to tokenize. Better not to input sentence in other languages
            since this class is mainly used for Chinese Word Segmentation.

        Returns
        -------
        ret : list of strs
            List of tokens
        """
        # we use default cutting mode provided by jieba, i.e., accurate mode
        return [
            tok for tok in self._tokenizer.cut(sample)
            if tok not in (' ', '')
        ]


class NLTKStanfordSegmenter:
    r"""Apply the Stanford Chinese Word Segmenter implemented in NLTK.

    Users of this class are required to install Java, NLTK and download Stanford Word Segmenter

    Parameters
    ----------
    segmenter_root : str, default '$MXNET_HOME/stanford-segmenter'
        Path to folder for storing stanford segmenter.
        MXNET_HOME defaults to '~/.mxnet'.

    slf4j_root : str, default '$MXNET_HOME/slf4j'
        Path to foler for storing slf4j.
        MXNET_HOME defaults to '~/.mxnet'

    java_class : str, default 'edu.stanford.nlp.ie.crf.CRFClassifier'
        The learning algorithm used for segmentation

    Examples
    --------
    >>> tokenizer = gluonnlp.data.NLTKStanfordSegmenter() #doctest:+SKIP
    >>> tokenizer('我来到北京清华大学') #doctest:+SKIP
    ['我', '来到', '北京', '清华大学']
    >>> tokenizer('小明硕士毕业于中国科学院计算所，后在日本京都大学深造') #doctest:+SKIP
    ['小明', '硕士', '毕业', '于', '中国科学院', '计算所', '，', '后', '在', '日本京都大学', '深造']

    """

    def __init__(self,
                 segmenter_root=os.path.join(get_home_dir(),
                                             'stanford-segmenter'),
                 slf4j_root=os.path.join(get_home_dir(), 'slf4j'),
                 java_class='edu.stanford.nlp.ie.crf.CRFClassifier'):
        is_java_exist = os.system('java -version')
        assert is_java_exist == 0, 'Java is not installed. You must install Java 8.0' \
                                   'in order to use the NLTKStanfordSegmenter'
        try:
            from nltk.tokenize import StanfordSegmenter  # pylint: disable=import-outside-toplevel
        except ImportError:
            raise ImportError(
                'NLTK or relevant packages are not installed. You must install NLTK '
                'in order to use the NLTKStanfordSegmenter. You can refer to the '
                'official installation guide in https://www.nltk.org/install.html.'
            )
        path_to_jar = os.path.join(segmenter_root,
                                   'stanford-segmenter-2018-02-27',
                                   'stanford-segmenter-3.9.1.jar')
        path_to_model = os.path.join(segmenter_root,
                                     'stanford-segmenter-2018-02-27', 'data',
                                     'pku.gz')
        path_to_dict = os.path.join(segmenter_root,
                                    'stanford-segmenter-2018-02-27', 'data',
                                    'dict-chris6.ser.gz')
        path_to_sihan_corpora_dict = os.path.join(
            segmenter_root, 'stanford-segmenter-2018-02-27', 'data')
        segmenter_url = 'https://nlp.stanford.edu/software/stanford-segmenter-2018-02-27.zip'
        segmenter_sha1 = 'aa27a6433704b7b4c6a14be1c126cb4b14b8f57b'
        stanford_segmenter = os.path.join(segmenter_root,
                                          'stanford-segmenter-2018-02-27.zip')
        if not os.path.exists(path_to_jar) or \
                not os.path.exists(path_to_model) or \
                not os.path.exists(path_to_dict) or \
                not os.path.exists(path_to_sihan_corpora_dict) or \
                not check_sha1(filename=stanford_segmenter, sha1_hash=segmenter_sha1):
            # automatically download the files from the website and place them to stanford_root
            if not os.path.exists(segmenter_root):
                os.mkdir(segmenter_root)
            download(url=segmenter_url,
                     path=segmenter_root,
                     sha1_hash=segmenter_sha1)
            _extract_archive(file=stanford_segmenter,
                             target_dir=segmenter_root)

        path_to_slf4j = os.path.join(slf4j_root, 'slf4j-1.7.25',
                                     'slf4j-api-1.7.25.jar')
        slf4j_url = 'https://www.slf4j.org/dist/slf4j-1.7.25.zip'
        slf4j_sha1 = '89ea41ad6ebe1b190139421bb7c8d981e9df1625'
        slf4j = os.path.join(slf4j_root, 'slf4j-1.7.25.zip')
        if not os.path.exists(path_to_slf4j) or \
                not check_sha1(filename=slf4j, sha1_hash=slf4j_sha1):
            # automatically download the files from the website and place them to slf4j_root
            if not os.path.exists(slf4j_root):
                os.mkdir(slf4j_root)
            download(url=slf4j_url, path=slf4j_root, sha1_hash=slf4j_sha1)
            _extract_archive(file=slf4j, target_dir=slf4j_root)
        self._tokenizer = StanfordSegmenter(
            java_class=java_class,
            path_to_jar=path_to_jar,
            path_to_slf4j=path_to_slf4j,
            path_to_dict=path_to_dict,
            path_to_sihan_corpora_dict=path_to_sihan_corpora_dict,
            path_to_model=path_to_model)

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The Chinese sentence to tokenize. Better not to input sentence in other languages
            since this class is mainly used for Chinese Word Segmentation.

        Returns
        -------
        ret : list of strs
            List of tokens
        """
        return self._tokenizer.segment(sample).strip().split()


class _SentencepieceProcessor:
    def __init__(self, path):
        try:
            import sentencepiece  # pylint: disable=import-outside-toplevel
        except ImportError:
            raise ImportError(
                'sentencepiece is not installed. You must install sentencepiece '
                'in order to use the Sentencepiece tokenizer and detokenizer. '
                'You can refer to the official installation guide '
                'in https://github.com/google/sentencepiece#installation')
        self._processor = sentencepiece.SentencePieceProcessor()
        self._processor.Load(path)

    def __len__(self):
        return len(self._processor)

    @property
    def tokens(self):
        return [self._processor.id_to_piece(i) for i in range(len(self))]


class SentencepieceTokenizer(_SentencepieceProcessor):
    r"""Apply the Sentencepiece Tokenizer, which supports subword tokenization such as BPE.

    Users of this class are required to `install sentencepiece
    <https://github.com/google/sentencepiece>`_. For example, one can use
    :samp:`pip install sentencepiece`


    Parameters
    ----------
    path : str
        Path to the pre-trained subword tokenization model.
    num_best : int, default 0
        A scalar for sampling subwords. If num_best = {0,1}, no sampling is performed.
        If num_best > 1, then samples from the num_best results.
        If num_best < 0, then assume that num_best is infinite and
        samples from the all hypothesis (lattice) using forward-filtering-and-backward-sampling
        algorithm.
    alpha : float, default 1.0
        A scalar for a smoothing parameter. Inverse temperature for probability rescaling.

    Examples
    --------
    >>> url = 'http://repo.mxnet.io/gluon/dataset/vocab/test-0690baed.bpe'
    >>> f = gluon.utils.download(url)
    -etc-
    >>> tokenizer = gluonnlp.data.SentencepieceTokenizer(f)
    >>> detokenizer = gluonnlp.data.SentencepieceDetokenizer(f)
    >>> sentence = 'This is a very awesome, life-changing sentence.'
    >>> tokenizer(sentence)
    ['▁This', '▁is', '▁a', '▁very', '▁awesome', ',', '▁life', '-', 'ch', 'anging', '▁sentence', '.']
    >>> detokenizer(tokenizer(sentence))
    'This is a very awesome, life-changing sentence.'
    >>> os.remove('test-0690baed.bpe')

    """

    def __init__(self, path, num_best=0, alpha=1.0):
        super(SentencepieceTokenizer, self).__init__(path)
        self._path = path
        self._nbest = num_best
        self._alpha = alpha

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The string to tokenize.

        Returns
        -------
        ret : list of strs
            List of tokens
        """
        return self._processor.SampleEncodeAsPieces(sample, self._nbest,
                                                    self._alpha)


class SentencepieceDetokenizer(_SentencepieceProcessor):
    r"""Apply the Sentencepiece detokenizer, which supports recombining subwords such as BPE.

    Users of this class are required to `install sentencepiece
    <https://github.com/google/sentencepiece>`_. For example, one can use
    :samp:`pip install sentencepiece`


    Parameters
    ----------
    path : str
        Path to the pre-trained subword tokenization model.

    Examples
    --------
    >>> url = 'http://repo.mxnet.io/gluon/dataset/vocab/test-0690baed.bpe'
    >>> f = gluon.utils.download(url)
    -etc-
    >>> tokenizer = gluonnlp.data.SentencepieceTokenizer(f)
    >>> detokenizer = gluonnlp.data.SentencepieceDetokenizer(f)
    >>> sentence = 'This is a very awesome, life-changing sentence.'
    >>> tokenizer(sentence)
    ['▁This', '▁is', '▁a', '▁very', '▁awesome', ',', '▁life', '-', 'ch', 'anging', '▁sentence', '.']
    >>> detokenizer(tokenizer(sentence))
    'This is a very awesome, life-changing sentence.'
    >>> os.remove('test-0690baed.bpe')

    """

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: list(str)
            The sentence to detokenize

        Returns
        -------
        ret : str
            Detokenized text
        """
        return self._processor.DecodePieces(sample)


class BERTBasicTokenizer:
    r"""Runs basic tokenization

    performs invalid character removal (e.g. control chars) and whitespace.
    tokenize CJK chars.
    splits punctuation on a piece of text.
    strips accents and convert to lower case.(If lower is true)

    Parameters
    ----------
    lower : bool, default True
        whether the text strips accents and convert to lower case.

    Examples
    --------
    >>> tokenizer = gluonnlp.data.BERTBasicTokenizer(lower=True)
    >>> tokenizer(' \tHeLLo!how  \n Are yoU?  ')
    ['hello', '!', 'how', 'are', 'you', '?']
    >>> tokenizer = gluonnlp.data.BERTBasicTokenizer(lower=False)
    >>> tokenizer(' \tHeLLo!how  \n Are yoU?  ')
    ['HeLLo', '!', 'how', 'Are', 'yoU', '?']

    """

    def __init__(self, lower=True):
        self.tokenizer = BasicTokenizer(lower=lower)

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample:  str
            The string to tokenize. Must be unicode.

        Returns
        -------
        ret : list of strs
            List of tokens
        """
        return self.tokenizer.tokenize(sample)

    def _is_control(self, char):
        """Checks whether `chars` is a control character."""
        return is_control(char, unicodedata.category(char))

    def _is_punctuation(self, char):
        """Checks whether `chars` is a punctuation character."""
        return is_punctuation(char, unicodedata.category(char))

    def _is_whitespace(self, char):
        """Checks whether `chars` is a whitespace character."""
        return is_whitespace(char, unicodedata.category(char))


class BERTTokenizer:
    r"""End-to-end tokenization for BERT models.

    Parameters
    ----------
    vocab
        Vocabulary for the corpus.
    lower
        whether the text strips accents and convert to lower case.
        If you use the BERT pre-training model,
        lower is set to Flase when using the cased model,
        otherwise it is set to True.
    max_input_chars_per_word
    lru_cache_size
        Maximum size of a least-recently-used cache to speed up tokenization.
        Use size of 2**20 for example.

    Examples
    --------
    >>> _, vocab = gluonnlp.model.bert_12_768_12(dataset_name='wiki_multilingual_uncased',
    ...                                          pretrained=False, root='./model')
    -etc-
    >>> tokenizer = gluonnlp.data.BERTTokenizer(vocab=vocab)
    >>> tokenizer('gluonnlp: 使NLP变得简单。')
    ['gl', '##uo', '##nn', '##lp', ':', '使', 'nl', '##p', '变', '得', '简', '单', '。']

    """

    _special_prefix = '##'

    def __init__(self, vocab: Vocab, lower: bool = True, max_input_chars_per_word: int = 200,
                 lru_cache_size: Optional[int] = None):
        self.vocab = vocab
        self.max_input_chars_per_word = max_input_chars_per_word
        self.basic_tokenizer = BasicTokenizer(lower=lower)
        self.wordpiece_tokenizer = WordpieceTokenizer(vocab=vocab,
                                                      unk_token=vocab.unknown_token,
                                                      max_input_chars_per_word=\
                                                              max_input_chars_per_word)

        if lru_cache_size:
            self._word_to_wordpiece_optimized = functools.lru_cache(maxsize=lru_cache_size)(
                self._word_to_wordpiece_optimized)

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The string to tokenize.

        Returns
        -------
        ret : list of strs
            List of tokens
        """

        return self._tokenizer(sample)

    def _tokenizer(self, text):
        split_tokens = []
        for token in self.basic_tokenizer.tokenize(text):
            for sub_token in self._word_to_wordpiece_optimized(token):
                split_tokens.append(sub_token)

        return split_tokens

    def _word_to_wordpiece_optimized(self, text):  # pylint: disable=method-hidden
        return self.wordpiece_tokenizer.tokenize(text)

    def convert_tokens_to_ids(self, tokens):
        """Converts a sequence of tokens into ids using the vocab."""
        return self.vocab.to_indices(tokens)

    @staticmethod
    def is_first_subword(token):
        """Check if a token is the beginning of subwords.

        Parameters
        ----------
        token : str
            The input token.

        Returns
        -------
        ret : True if the token is the beginning of a serious of wordpieces.

        Examples
        --------
        >>> _, vocab = gluonnlp.model.bert_12_768_12(dataset_name='wiki_multilingual_uncased',
        ...                                          pretrained=False, root='./bert_tokenizer')
        -etc-
        >>> tokenizer = gluonnlp.data.BERTTokenizer(vocab=vocab)
        >>> tokenizer('gluonnlp: 使NLP变得简单。')
        ['gl', '##uo', '##nn', '##lp', ':', '使', 'nl', '##p', '变', '得', '简', '单', '。']
        >>> tokenizer.is_first_subword('gl')
        True
        >>> tokenizer.is_first_subword('##uo')
        False
        """
        return not token.startswith(BERTTokenizer._special_prefix)


class BERTSPTokenizer:
    r"""End-to-end SentencePiece tokenization for BERT models.

    It works best with BERTSentenceTransform().

    .. note::

        BERTSPTokenizer depends on the sentencepiece library. For multi-processing
        with BERTSPTokenizer, making an extra copy of the BERTSPTokenizer instance
        is recommended before using it.

    Parameters
    ----------
    path : str
        Path to the pre-trained subword tokenization model.
    vocab : gluonnlp.Vocab
        Vocabulary for the corpus.
    num_best : int, default 0
        A scalar for sampling subwords. If num_best = {0,1}, no sampling is performed.
        If num_best > 1, then samples from the num_best results.
        If num_best < 0, then assume that num_best is infinite and
        samples from the all hypothesis (lattice) using forward-filtering-and-backward-sampling
        algorithm.
    alpha : float
        A scalar for a smoothing parameter. Inverse temperature for probability rescaling.
    lower : bool, default True
        Whether the text strips accents and convert to lower case.
        If you use the BERT pre-training model,
        lower is set to False when using the cased model,
        otherwise it is set to True.
    max_input_chars_per_word : int, default 200

    Examples
    --------
    >>> url = 'http://repo.mxnet.io/gluon/dataset/vocab/test-682b5d15.bpe'
    >>> f = gluon.utils.download(url)
    -etc-
    >>> bert_vocab = gluonnlp.vocab.BERTVocab.from_sentencepiece(f)
    >>> sp_tokenizer = BERTSPTokenizer(f, bert_vocab, lower=True)
    >>> sentence = 'Better is to bow than break.'
    >>> sp_tokenizer(sentence)
    ['▁better', '▁is', '▁to', '▁b', 'ow', '▁than', '▁brea', 'k', '▁', '.']
    >>> os.remove('test-682b5d15.bpe')
    """

    _special_prefix = '▁'

    def __init__(self,
                 path,
                 vocab,
                 num_best=0,
                 alpha=1.0,
                 lower=True,
                 max_input_chars_per_word=200):
        self._path = path
        self._num_best = num_best
        self._alpha = alpha
        self.sentencepiece = None
        self.basic_tokenizer = BERTBasicTokenizer(lower=lower)
        self.vocab = vocab
        self.max_input_chars_per_word = max_input_chars_per_word


    def _activate_sp(self):
        self.sentencepiece = SentencepieceTokenizer(self._path, self._num_best,
                                                    self._alpha)

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample: str
            The string to tokenize.

        Returns
        -------
        ret : list of strs
            List of tokens
        """

        return self._tokenizer(sample)

    def _tokenizer(self, text):
        split_tokens = []
        for token in self.basic_tokenizer(text):
            for sub_token in self._tokenize_wordpiece(token):
                split_tokens.append(sub_token)

        return split_tokens

    def _tokenize_wordpiece(self, text):
        """Tokenizes a piece of text into its word pieces.

        This use Google's SentencePiece tokenizer model file

        For example:
          input = "unaffable"
          output = ["▁un", "aff", "able"]

        Args:
          text: A single token or whitespace separated tokens. This should have
            already been passed through `BERTBasicTokenizer.

        Returns:
          A list of sentencepieced tokens.
        """
        # Swig object can not be pickled when multiprocessing.
        if self.sentencepiece is None:
            self._activate_sp()
        output_tokens = self.sentencepiece(text)
        return output_tokens

    def convert_tokens_to_ids(self, tokens):
        """Converts a sequence of tokens into ids using the vocab."""
        return self.vocab.to_indices(tokens)

    @staticmethod
    def is_first_subword(token):
        """Check if a string token is a subword following a previous subword,
        instead of the beginning of a word.

        Parameters
        ----------
        token : str
            The input token.

        Returns
        -------
        ret : True if the token is the beginning of a series of subwords,

        Examples
        --------
        >>> url = 'http://repo.mxnet.io/gluon/dataset/vocab/test-682b5d15.bpe'
        >>> f = gluon.utils.download(url)
        -etc-
        >>> bert_vocab = gluonnlp.vocab.BERTVocab.from_sentencepiece(f)
        >>> sp_tokenizer = BERTSPTokenizer(f, bert_vocab, lower=True)
        >>> sp_tokenizer('Better is to bow than break.')
        ['▁better', '▁is', '▁to', '▁b', 'ow', '▁than', '▁brea', 'k', '▁', '.']
        >>> sp_tokenizer.is_first_subword('▁better')
        True
        >>> sp_tokenizer.is_first_subword('ow')
        False
        >>> os.remove('test-682b5d15.bpe')
        """
        return token.startswith(BERTSPTokenizer._special_prefix)


class BERTSentenceTransform:
    r"""BERT style data transformation.

    Parameters
    ----------
    tokenizer : BERTTokenizer.
        Tokenizer for the sentences.
    max_seq_length : int.
        Maximum sequence length of the sentences.
    vocab : Vocab
        The vocabulary which has cls_token and sep_token registered.
        If vocab.cls_token is not present, vocab.bos_token is used instead.
        If vocab.sep_token is not present, vocab.eos_token is used instead.
    pad : bool, default True
        Whether to pad the sentences to maximum length.
    pair : bool, default True
        Whether to transform sentences or sentence pairs.
    """

    def __init__(self, tokenizer, max_seq_length, vocab=None, pad=True, pair=True):
        self._tokenizer = tokenizer
        self._max_seq_length = max_seq_length
        self._pad = pad
        self._pair = pair
        self._vocab = self._tokenizer.vocab if vocab is None else vocab
        # RoBERTa does not register CLS token and SEP token
        if hasattr(self._vocab, 'cls_token'):
            self._cls_token = self._vocab.cls_token
        else:
            self._cls_token = self._vocab.bos_token
        if hasattr(self._vocab, 'sep_token'):
            self._sep_token = self._vocab.sep_token
        else:
            self._sep_token = self._vocab.eos_token
        self._padding_token = self._vocab.padding_token

    def __call__(self, line):
        """Perform transformation for sequence pairs or single sequences.

        The transformation is processed in the following steps:
        - tokenize the input sequences
        - insert [CLS], [SEP] as necessary
        - generate type ids to indicate whether a token belongs to the first
        sequence or the second sequence.
        - generate valid length

        For sequence pairs, the input is a tuple of 2 strings:
        text_a, text_b.

        Inputs:
            text_a: 'is this jacksonville ?'
            text_b: 'no it is not'
        Tokenization:
            text_a: 'is this jack ##son ##ville ?'
            text_b: 'no it is not .'
        Processed:
            tokens: '[CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]'
            type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
            valid_length: 14

        For single sequences, the input is a tuple of single string:
        text_a.

        Inputs:
            text_a: 'the dog is hairy .'
        Tokenization:
            text_a: 'the dog is hairy .'
        Processed:
            text_a: '[CLS] the dog is hairy . [SEP]'
            type_ids: 0     0   0   0  0     0 0
            valid_length: 7

        If vocab.cls_token and vocab.sep_token are not present,
        vocab.bos_token and vocab.eos_token are used instead.

        Parameters
        ----------
        line: tuple of str
            Input strings. For sequence pairs, the input is a tuple of 2 strings:
            (text_a, text_b). For single sequences, the input is a tuple of single
            string: (text_a,).

        Returns
        -------
        np.array: input token ids in 'int32', shape (batch_size, seq_length)
        np.array: valid length in 'int32', shape (batch_size,)
        np.array: input token type ids in 'int32', shape (batch_size, seq_length)

        """

        # convert to unicode
        text_a = line[0]
        if self._pair:
            assert len(line) == 2
            text_b = line[1]

        tokens_a = self._tokenizer(text_a)
        tokens_b = None

        if self._pair:
            tokens_b = self._tokenizer(text_b)

        if tokens_b:
            # Modifies `tokens_a` and `tokens_b` in place so that the total
            # length is less than the specified length.
            # Account for [CLS], [SEP], [SEP] with "- 3"
            self._truncate_seq_pair(tokens_a, tokens_b,
                                    self._max_seq_length - 3)
        else:
            # Account for [CLS] and [SEP] with "- 2"
            if len(tokens_a) > self._max_seq_length - 2:
                tokens_a = tokens_a[0:(self._max_seq_length - 2)]

        # The embedding vectors for `type=0` and `type=1` were learned during
        # pre-training and are added to the wordpiece embedding vector
        # (and position vector). This is not *strictly* necessary since
        # the [SEP] token unambiguously separates the sequences, but it makes
        # it easier for the model to learn the concept of sequences.

        # For classification tasks, the first vector (corresponding to [CLS]) is
        # used as as the "sentence vector". Note that this only makes sense because
        # the entire model is fine-tuned.
        tokens = []
        tokens.append(self._cls_token)
        tokens.extend(tokens_a)
        tokens.append(self._sep_token)
        segment_ids = [0] * len(tokens)

        if tokens_b:
            tokens.extend(tokens_b)
            tokens.append(self._sep_token)
            segment_ids.extend([1] * (len(tokens) - len(segment_ids)))

        input_ids = self._vocab[tokens]

        # The valid length of sentences. Only real  tokens are attended to.
        valid_length = len(input_ids)

        if self._pad:
            # Zero-pad up to the sequence length.
            padding_length = self._max_seq_length - valid_length
            # use padding tokens for the rest
            input_ids.extend([self._vocab[self._padding_token]] * padding_length)
            segment_ids.extend([0] * padding_length)

        return np.array(input_ids, dtype='int32'), np.array(valid_length, dtype='int32'),\
            np.array(segment_ids, dtype='int32')

    def _truncate_seq_pair(self, tokens_a, tokens_b, max_length):
        """Truncates a sequence pair in place to the maximum length."""
        # This is a simple heuristic which will always truncate the longer sequence
        # one token at a time. This makes more sense than truncating an equal percent
        # of tokens from each, since if one sequence is very short then each token
        # that's truncated likely contains more information than a longer sequence.
        while True:
            total_length = len(tokens_a) + len(tokens_b)
            if total_length <= max_length:
                break
            if len(tokens_a) > len(tokens_b):
                tokens_a.pop()
            else:
                tokens_b.pop()

class _GPT2BPE:
    """Base class for GPT-2 BPE tokenizer and detokenizer."""
    def __init__(self):
        codes = list(range(ord('!'), ord('~') + 1)) +\
                list(range(ord('¡'), ord('¬') + 1)) +\
                list(range(ord('®'), ord('ÿ') + 1))
        chr_fn = chr
        try:
            chr_fn(256)
        except ValueError:
            chr_fn = unichr # noqa: F821
        byte_encoder = {code: chr_fn(code) for code in codes}
        shift = 0
        for code in range(2 ** 8):
            if code not in byte_encoder:
                byte_encoder[code] = chr_fn(2 ** 8 + shift)
                shift += 1
        self._byte_encoder = byte_encoder


class GPT2BPETokenizer(_GPT2BPE):
    """BPE tokenizer used in OpenAI GPT-2 model.

    Parameters
    ----------
    root : str, default '$MXNET_HOME/models'
        Location for keeping the BPE rank file.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    bpe_ranks_file_hash = ('openai_webtext_bpe_ranks-396d4d8e.json',
                           '396d4d8ec90cb02f4d56e049e0e4add868bcd943')
    bpe_ranks_archive_hash = ('openai_webtext_bpe_ranks-396d4d8e.zip',
                              '1a770728fd102bc9dc332f322e6bfb294767a685')
    def __init__(self, root=os.path.join(get_home_dir(), 'models')):
        try:
            import regex  # pylint: disable=import-outside-toplevel
            self._regex = regex
        except ImportError:
            raise ImportError(
                'GPT2BPETokenizer requires regex. '
                'To install regex, use pip install -U regex')
        super(GPT2BPETokenizer, self).__init__()
        root = os.path.expanduser(root)
        file_name, sha1_hash = self.bpe_ranks_file_hash
        file_path = os.path.join(root, file_name)
        if not os.path.exists(file_path) or not check_sha1(file_path, sha1_hash):
            if os.path.exists(file_path):
                print('Detected mismatch in the content of BPE rank file. Downloading again.')
            else:
                print('BPE rank file is not found. Downloading.')
            if not os.path.exists(root):
                try:
                    os.makedirs(root)
                except OSError as e:
                    if e.errno == errno.EEXIST and os.path.isdir(root):
                        pass
                    else:
                        raise e

            prefix = str(time.time())
            zip_file_path = os.path.join(root, prefix + file_name)
            repo_url = _get_repo_url()
            if repo_url[-1] != '/':
                repo_url = repo_url + '/'
            archive_name, archive_hash = self.bpe_ranks_archive_hash
            _url_format = '{repo_url}gluon/dataset/vocab/{file_name}'
            download(_url_format.format(repo_url=repo_url, file_name=archive_name),
                     path=zip_file_path,
                     sha1_hash=archive_hash,
                     overwrite=True)
            with zipfile.ZipFile(zip_file_path) as zf:
                if not os.path.exists(file_path):
                    zf.extractall(root)
            try:
                os.remove(zip_file_path)
            except OSError as e:
                # file has already been removed.
                if e.errno == 2:
                    pass
                else:
                    raise e

            if not check_sha1(file_path, sha1_hash):
                raise ValueError('Downloaded file has different hash. Please try again.')
        self._read_bpe_ranks(file_path)
        self._cache = {}
        self._token_pattern = self._regex.compile(
            r'\'s|\'t|\'re|\'ve|\'m|\'ll|\'d| ?\p{L}+'
            r'| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+')

    def _read_bpe_ranks(self, file_path):
        with io.open(file_path, 'r', encoding='utf-8') as f:
            bpe_data = f.read()
            self._bpe_ranks = {
                tuple(merge_str.split()): i for i, merge_str
                in enumerate(bpe_data.split('\n')[1:-1])}

    def get_bpe_subword(self, token):
        """ Encode the word token into BPE subwords

        Parameters
        ----------
        token : str

        Returns
        -------
        chars : list(str)
        """
        if token in self._cache:
            return self._cache[token]
        chars = list(token)
        while len(chars) > 0:
            min_pair, min_rank = None, float('inf')
            # Find the pair with the minimum rank
            for i in range(1, len(chars)):
                pair = (chars[i - 1], chars[i])
                rank = self._bpe_ranks.get(pair, float('inf'))
                if rank < min_rank:
                    min_rank = rank
                    min_pair = pair
            if min_pair is None or min_pair not in self._bpe_ranks:
                break
            # Merge the pair
            last, tail = chars[0], 1
            for index in range(1, len(chars)):
                if (last, chars[index]) == min_pair:
                    chars[tail - 1] = last + chars[index]
                    last = last + chars[index]
                else:
                    chars[tail - 1] = last
                    tail += 1
                    last = chars[index]
            chars[tail - 1] = last
            chars = chars[:tail]
        self._cache[token] = chars
        return chars

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample : str

        Returns
        -------
        ret : list(str)
        """
        ret = []
        for word_token in self._regex.findall(self._token_pattern, sample):
            word_token = bytearray(word_token.encode('utf-8'))
            word_token = ''.join(self._byte_encoder[code] for code in word_token)
            ret.extend(self.get_bpe_subword(word_token))
        return ret


class GPT2BPEDetokenizer(_GPT2BPE):
    """BPE detokenizer used in OpenAI GPT-2 model."""
    def __init__(self):
        super(GPT2BPEDetokenizer, self).__init__()
        self._byte_decoder = {v: k for k, v in self._byte_encoder.items()}

    def __call__(self, sample):
        """

        Parameters
        ----------
        sample : list(str)

        Returns
        -------
        ret : str
        """
        text = ''.join(sample)
        ret = bytearray(
            [self._byte_decoder[byte] for byte in text]).decode('utf-8', errors='replace')
        return ret



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Utility classes and functions. They help organize and keep statistics of datasets."""
import collections
import os
import tarfile
import zipfile
import random
import sys
import shutil

import numpy as np
from mxnet.gluon.data import SimpleDataset
from mxnet.gluon.utils import _get_repo_url, check_sha1, download

from .. import _constants as C
from .. import utils

__all__ = [
    'Counter', 'count_tokens', 'concat_sequence', 'slice_sequence', 'train_valid_split',
    'line_splitter', 'whitespace_splitter', 'Splitter'
]


class Counter(collections.Counter):  # pylint: disable=abstract-method
    """Counter class for keeping token frequencies."""

    def discard(self, min_freq, unknown_token):
        """Discards tokens with frequency below min_frequency and represents them
        as `unknown_token`.

        Parameters
        ----------
        min_freq: int
            Tokens whose frequency is under min_freq is counted as `unknown_token` in
            the Counter returned.
        unknown_token: str
            The representation for any unknown token.

        Returns
        -------
        The Counter instance.

        Examples
        --------
        >>> a = gluonnlp.data.Counter({'a': 10, 'b': 1, 'c': 1})
        >>> a.discard(3, '<unk>')
        Counter({'a': 10, '<unk>': 2})
        """
        freq = 0
        ret = Counter({})
        for token, count in self.items():
            if count < min_freq:
                freq += count
            else:
                ret[token] = count
        ret[unknown_token] = ret.get(unknown_token, 0) + freq
        return ret


class DefaultLookupDict(dict):
    """Dictionary class with fall-back look-up with default value set in the constructor."""

    def __init__(self, default, d=None):
        if d:
            super(DefaultLookupDict, self).__init__(d)
        else:
            super(DefaultLookupDict, self).__init__()
        self._default = default

    def __getitem__(self, k):
        return self.get(k, self._default)


def count_tokens(tokens, to_lower=False, counter=None):
    r"""Counts tokens in the specified string.

    For token_delim='(td)' and seq_delim='(sd)', a specified string of two sequences of tokens may
    look like::

        (td)token1(td)token2(td)token3(td)(sd)(td)token4(td)token5(td)(sd)


    Parameters
    ----------
    tokens : list of str
        A source list of tokens.
    to_lower : bool, default False
        Whether to convert the source source_str to the lower case.
    counter : Counter or None, default None
        The Counter instance to be updated with the counts of `tokens`. If
        None, return a new Counter instance counting tokens from `tokens`.

    Returns
    -------
    The `counter` Counter instance after being updated with the token
    counts of `source_str`. If `counter` is None, return a new Counter
    instance counting tokens from `source_str`.

    Examples
    --------
    >>> import re
    >>> source_str = ' Life is great ! \n life is good . \n'
    >>> source_str_tokens = filter(None, re.split(' |\n', source_str))
    >>> counter = gluonnlp.data.count_tokens(source_str_tokens)
    >>> sorted(counter.items())
    [('!', 1), ('.', 1), ('Life', 1), ('good', 1), ('great', 1), ('is', 2), ('life', 1)]

    """
    if to_lower:
        tokens = [t.lower() for t in tokens]

    if counter is None:
        return Counter(tokens)
    else:
        counter.update(tokens)
        return counter


def concat_sequence(sequences):
    """Concatenate sequences of tokens into a single flattened list of tokens.

    Parameters
    ----------
    sequences : list of list of object
        Sequences of tokens, each of which is an iterable of tokens.

    Returns
    -------
    Flattened list of tokens.

    """
    return [token for seq in sequences for token in seq if token]


def slice_sequence(sequence, length, pad_last=False, pad_val=C.PAD_TOKEN, overlap=0):
    """Slice a flat sequence of tokens into sequences tokens, with each
    inner sequence's length equal to the specified `length`, taking into account the requested
    sequence overlap.

    Parameters
    ----------
    sequence : list of object
        A flat list of tokens.
    length : int
        The length of each of the samples.
    pad_last : bool, default False
        Whether to pad the last sequence when its length doesn't align. If the last sequence's
        length doesn't align and ``pad_last`` is False, it will be dropped.
    pad_val : object, default
        The padding value to use when the padding of the last sequence is enabled. In general,
        the type of ``pad_val`` should be the same as the tokens.
    overlap : int, default 0
        The extra number of items in current sample that should overlap with the
        next sample.

    Returns
    -------
    List of list of tokens, with the length of each inner list equal to `length`.

    """
    if length <= overlap:
        raise ValueError('length needs to be larger than overlap')

    if pad_last:
        pad_len = _slice_pad_length(len(sequence), length, overlap)
        sequence = sequence + [pad_val] * pad_len
    num_samples = (len(sequence) - length) // (length - overlap) + 1

    return [sequence[i * (length - overlap): ((i + 1) * length - i * overlap)]
            for i in range(num_samples)]


def _slice_pad_length(num_items, length, overlap=0):
    """Calculate the padding length needed for sliced samples in order not to discard data.

    Parameters
    ----------
    num_items : int
        Number of items in dataset before collating.
    length : int
        The length of each of the samples.
    overlap : int, default 0
        The extra number of items in current sample that should overlap with the
        next sample.

    Returns
    -------
    Length of paddings.

    """
    if length <= overlap:
        raise ValueError('length needs to be larger than overlap')

    step = length - overlap
    span = num_items - length
    residual = span % step
    if residual:
        return step - residual
    else:
        return 0


# name:[sha hash, file extension, special tokens]
_vocab_sha1 = {'wikitext-2':
               ['be36dc5238c2e7d69720881647ab72eb506d0131', '.vocab', {}],
               'gbw':
               ['ebb1a287ca14d8fa6f167c3a779e5e7ed63ac69f', '.vocab', {}],
               'WMT2014_src':
               ['230ebb817b1d86950d71e2e765f192a4e4f34415', '.vocab', {}],
               'WMT2014_tgt':
               ['230ebb817b1d86950d71e2e765f192a4e4f34415', '.vocab', {}],
               'book_corpus_wiki_en_cased':
               ['2d62af22535ed51f35cc8e2abb607723c89c2636', '.vocab', {}],
               'book_corpus_wiki_en_uncased':
               ['a66073971aa0b1a262453fe51342e57166a8abcf', '.vocab', {}],
               'openwebtext_book_corpus_wiki_en_uncased':
               ['a66073971aa0b1a262453fe51342e57166a8abcf', '.vocab', {}],
               'openwebtext_ccnews_stories_books_cased':
               ['2b804f8f90f9f93c07994b703ce508725061cf43', '.vocab', {}],
               'wiki_multilingual_cased':
               ['0247cb442074237c38c62021f36b7a4dbd2e55f7', '.vocab', {}],
               'distilbert_book_corpus_wiki_en_uncased':
               ['80ef760a6bdafec68c99b691c94ebbb918c90d02', '.vocab', {}],
               'wiki_cn_cased':
               ['ddebd8f3867bca5a61023f73326fb125cf12b4f5', '.vocab', {}],
               'wiki_multilingual_uncased':
               ['2b2514cc539047b9179e9d98a4e68c36db05c97a', '.vocab', {}],
               'scibert_scivocab_uncased':
               ['2d2566bfc416790ab2646ab0ada36ba628628d60', '.vocab', {}],
               'scibert_scivocab_cased':
               ['2c714475b521ab8542cb65e46259f6bfeed8041b', '.vocab', {}],
               'scibert_basevocab_uncased':
               ['80ef760a6bdafec68c99b691c94ebbb918c90d02', '.vocab', {}],
               'scibert_basevocab_cased':
               ['a4ff6fe1f85ba95f3010742b9abc3a818976bb2c', '.vocab', {}],
               'biobert_v1.0_pmc_cased':
               ['a4ff6fe1f85ba95f3010742b9abc3a818976bb2c', '.vocab', {}],
               'biobert_v1.0_pubmed_cased':
               ['a4ff6fe1f85ba95f3010742b9abc3a818976bb2c', '.vocab', {}],
               'biobert_v1.0_pubmed_pmc_cased':
               ['a4ff6fe1f85ba95f3010742b9abc3a818976bb2c', '.vocab', {}],
               'biobert_v1.1_pubmed_cased':
               ['a4ff6fe1f85ba95f3010742b9abc3a818976bb2c', '.vocab', {}],
               'clinicalbert_uncased':
               ['80ef760a6bdafec68c99b691c94ebbb918c90d02', '.vocab', {}],
               'baidu_ernie_uncased':
               ['223553643220255e2a0d4c60e946f4ad7c719080', '.vocab', {}],
               'openai_webtext':
               ['f917dc7887ce996068b0a248c8d89a7ec27b95a1', '.vocab', {}],
               'xlnet_126gb':
               ['0d74490383bbc5c62b8bcea74d8b74a1bb1280b3', '.vocab', {}],
               'kobert_news_wiki_ko_cased':
               ['f86b1a8355819ba5ab55e7ea4a4ec30fdb5b084f', '.spiece', {'padding_token': '[PAD]'}]}

_url_format = '{repo_url}gluon/dataset/vocab/{file_name}.zip'


def train_valid_split(dataset, valid_ratio=0.05, stratify=None):
    """Split the dataset into training and validation sets.

    Parameters
    ----------
    dataset : list
        A list of training samples.
    valid_ratio : float, default 0.05
        Proportion of training samples to use for validation set
        range: [0, 1]
    stratify : list, default None
        If not None, data is split in a stratified fashion,
        using the contents of stratify as class labels.

    Returns
    -------
    train : SimpleDataset
    valid : SimpleDataset
    """
    if not 0.0 <= valid_ratio <= 1.0:
        raise ValueError('valid_ratio should be in [0, 1]')

    if not stratify:
        num_train = len(dataset)
        num_valid = np.ceil(num_train * valid_ratio).astype('int')
        indices = np.arange(num_train)

        np.random.shuffle(indices)
        valid = SimpleDataset([dataset[indices[i]] for i in range(num_valid)])
        train = SimpleDataset(
            [dataset[indices[i + num_valid]] for i in range(num_train - num_valid)])

        return train, valid
    else:
        if not isinstance(stratify, list):
            raise TypeError('stratify should be a list')
        if not len(stratify) == len(dataset):
            raise ValueError('stratify should be the same length as num_train')

        classes, digitized = np.unique(stratify, return_inverse=True)
        n_classes = len(classes)
        num_class = np.bincount(digitized)
        num_valid = np.ceil(valid_ratio * num_class).astype('int')

        valid = []
        train = []

        for idx in range(n_classes):
            indices = np.nonzero(stratify == classes[idx])[0]
            np.random.shuffle(indices)
            valid += [dataset[indices[i]] for i in range(num_valid[idx])]
            train += [dataset[indices[i + num_valid[idx]]]
                      for i in range(num_class[idx] - num_valid[idx])]

        np.random.shuffle(valid)
        np.random.shuffle(train)

        train = SimpleDataset(train)
        valid = SimpleDataset(valid)

        return train, valid


def short_hash(name):
    if name not in _vocab_sha1:
        vocabs = list(_vocab_sha1.keys())
        raise ValueError('Vocabulary for {name} is not available. '
                         'Hosted vocabularies include: {vocabs}'.format(name=name,
                                                                        vocabs=vocabs))
    return _vocab_sha1[name][0][:8]


def _get_vocab_tokenizer_info(name, root):
    file_name = '{name}-{short_hash}'.format(name=name,
                                             short_hash=short_hash(name))
    root = os.path.expanduser(root)
    sha1_hash, file_ext, special_tokens = _vocab_sha1[name]
    return file_name, file_ext, sha1_hash, special_tokens


def _download_vocab_tokenizer(root, file_name, file_ext, file_path):
    utils.mkdir(root)

    temp_num = str(random.Random().randint(1, sys.maxsize))
    temp_root = os.path.join(root, temp_num)
    temp_file_path = os.path.join(temp_root, file_name + file_ext)
    temp_zip_file_path = os.path.join(temp_root, temp_num + '_' + file_name + '.zip')

    repo_url = _get_repo_url()
    download(_url_format.format(repo_url=repo_url, file_name=file_name),
             path=temp_zip_file_path, overwrite=True)
    with zipfile.ZipFile(temp_zip_file_path) as zf:
        assert file_name + file_ext in zf.namelist(), '{} not part of {}. Only have: {}'.format(
            file_name + file_ext, file_name + '.zip', zf.namelist())
        utils.mkdir(temp_root)
        zf.extractall(temp_root)
        os.replace(temp_file_path, file_path)
        shutil.rmtree(temp_root)

def _load_pretrained_vocab(name, root, cls=None):
    """Load the accompanying vocabulary object for pre-trained model.

    Parameters
    ----------
    name : str
        Name of the vocabulary, usually the name of the dataset.
    root : str
        Location for keeping the model vocabulary.
    cls : nlp.Vocab or nlp.vocab.BERTVocab, default nlp.Vocab

    Returns
    -------
    Vocab or nlp.vocab.BERTVocab, Tokenizer or None
        Loaded vocabulary object and Tokenizer for the pre-trained model.
    """
    file_name, file_ext, sha1_hash, special_tokens = _get_vocab_tokenizer_info(name, root)
    file_path = os.path.join(root, file_name + file_ext)
    if os.path.exists(file_path):
        if check_sha1(file_path, sha1_hash):
            return _load_vocab_file(file_path, cls, **special_tokens)
        else:
            print('Detected mismatch in the content of model vocab file. Downloading again.')
    else:
        print('Vocab file is not found. Downloading.')
    _download_vocab_tokenizer(root, file_name, file_ext, file_path)
    if check_sha1(file_path, sha1_hash):
        return _load_vocab_file(file_path, cls, **special_tokens)
    else:
        raise ValueError('Downloaded file has different hash. Please try again.')


def _load_pretrained_sentencepiece_tokenizer(name, root, **kwargs):
    from ..data import SentencepieceTokenizer  # pylint: disable=import-outside-toplevel
    file_name, file_ext, sha1_hash, _ = _get_vocab_tokenizer_info(name, root)
    file_path = os.path.join(root, file_name + file_ext)
    if os.path.exists(file_path):
        if check_sha1(file_path, sha1_hash):
            assert file_path.endswith('.spiece')
            return SentencepieceTokenizer(file_path, **kwargs)
        else:
            print('Detected mismatch in the content of model tokenizer file. Downloading again.')
    else:
        print('tokenizer file is not found. Downloading.')
    _download_vocab_tokenizer(root, file_name, file_ext, file_path)
    if check_sha1(file_path, sha1_hash):
        assert file_path.endswith('.spiece')
        return SentencepieceTokenizer(file_path, **kwargs)
    else:
        raise ValueError('Downloaded file has different hash. Please try again.')


def _load_vocab_file(file_path, cls, **kwargs):
    with open(file_path, 'r') as f:
        if cls is None:
            from ..vocab import Vocab  # pylint: disable=import-outside-toplevel
            cls = Vocab
        if file_path.endswith('.spiece'):
            assert kwargs is not None, 'special tokens must be specified when .spiece provide.'
            from ..vocab import BERTVocab  # pylint: disable=import-outside-toplevel
            return BERTVocab.from_sentencepiece(
                file_path,
                **kwargs)
        else:
            return cls.from_json(f.read())


def _extract_archive(file, target_dir):  # pylint: disable=redefined-builtin
    """Extract archive file

    Parameters
    ----------
    file : str
        Absolute path of the archive file.
    target_dir : str
        Target directory of the archive to be uncompressed

    """
    if file.endswith('.gz') or file.endswith('.tar') or file.endswith('.tgz'):
        archive = tarfile.open(file, 'r')
    elif file.endswith('.zip'):
        archive = zipfile.ZipFile(file, 'r')
    else:
        raise Exception('Unrecognized file type: ' + file)
    archive.extractall(path=target_dir)
    archive.close()


def line_splitter(s):
    """Split a string at newlines.

    Parameters
    ----------
    s : str
        The string to be split

    Returns
    --------
    List[str]
        List of strings. Obtained by calling s.splitlines().

    """
    return s.splitlines()


def whitespace_splitter(s):
    """Split a string at whitespace (space, tab, newline, return, formfeed).

    Parameters
    ----------
    s : str
        The string to be split

    Returns
    --------
    List[str]
        List of strings. Obtained by calling s.split().
    """
    return s.split()


class Splitter:
    """Split a string based on a separator.

    Parameters
    ----------
    separator : str
        The separator based on which string is split.
    """

    def __init__(self, separator=None):
        self._separator = separator

    def __call__(self, s):
        """Split a string based on the separator.

        Parameters
        ----------
        s : str
            The string to be split

        Returns
        --------
        List[str]
            List of strings. Obtained by calling s.split(separator).
        """
        return s.split(self._separator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=
"""Language model RNN loss."""

__all__ = ['ActivationRegularizationLoss', 'TemporalActivationRegularizationLoss']

from mxnet.gluon.loss import Loss


class ActivationRegularizationLoss(Loss):
    r"""Computes Activation Regularization Loss. (alias: AR)

    The formulation is as below:

    .. math::

        L = \alpha L_2(h_t)

    where :math:`L_2(\cdot) = {||\cdot||}_2, h_t` is the output of the RNN at timestep t.
    :math:`\alpha` is scaling coefficient.

    The implementation follows the work::

        @article{merity2017revisiting,
          title={Revisiting Activation Regularization for Language RNNs},
          author={Merity, Stephen and McCann, Bryan and Socher, Richard},
          journal={arXiv preprint arXiv:1708.01009},
          year={2017}
        }

    Parameters
    ----------
    alpha : float, default 0
        The scaling coefficient of the regularization.
    weight : float or None
        Global scalar weight for loss.
    batch_axis : int, default 0
        The axis that represents mini-batch.
    """
    def __init__(self, alpha=0, weight=None, batch_axis=None, **kwargs):
        super(ActivationRegularizationLoss, self).__init__(weight, batch_axis, **kwargs)
        self._alpha = alpha

    def __repr__(self):
        s = 'ActivationRegularizationLoss (alpha={alpha})'
        return s.format(alpha=self._alpha)

    def hybrid_forward(self, F, *states): # pylint: disable=arguments-differ
        """
        Parameters
        ----------
        states : list
            the stack outputs from RNN, which consists of output from each time step (TNC).

        Returns
        --------
        loss : NDArray
            loss tensor with shape (batch_size,). Dimensions other than batch_axis are averaged out.
        """
        # pylint: disable=unused-argument
        if self._alpha != 0:
            if states:
                means = [self._alpha * state.__pow__(2).mean()
                         for state in states[-1:]]
                return F.add_n(*means)
            else:
                return F.zeros(1)
        return F.zeros(1)


class TemporalActivationRegularizationLoss(Loss):
    r"""Computes Temporal Activation Regularization Loss. (alias: TAR)

    The formulation is as below:

    .. math::

        L = \beta L_2(h_t-h_{t+1})

    where :math:`L_2(\cdot) = {||\cdot||}_2, h_t` is the output of the RNN at timestep t,
    :math:`h_{t+1}` is the output of the RNN at timestep t+1, :math:`\beta` is scaling coefficient.

    The implementation follows the work::

        @article{merity2017revisiting,
          title={Revisiting Activation Regularization for Language RNNs},
          author={Merity, Stephen and McCann, Bryan and Socher, Richard},
          journal={arXiv preprint arXiv:1708.01009},
          year={2017}
        }

    Parameters
    ----------
    beta : float, default 0
        The scaling coefficient of the regularization.
    weight : float or None
        Global scalar weight for loss.
    batch_axis : int, default 0
        The axis that represents mini-batch.
    """

    def __init__(self, beta=0, weight=None, batch_axis=None, **kwargs):
        super(TemporalActivationRegularizationLoss, self).__init__(weight, batch_axis, **kwargs)
        self._beta = beta

    def __repr__(self):
        s = 'TemporalActivationRegularizationLoss (beta={beta})'
        return s.format(beta=self._beta)

    def hybrid_forward(self, F, *states): # pylint: disable=arguments-differ
        """
        Parameters
        ----------
        states : list
            the stack outputs from RNN, which consists of output from each time step (TNC).

        Returns
        --------
        loss : NDArray
            loss tensor with shape (batch_size,). Dimensions other than batch_axis are averaged out.
        """
        # pylint: disable=unused-argument
        if self._beta != 0:
            if states:
                means = [self._beta * (state[1:] - state[:-1]).__pow__(2).mean()
                         for state in states[-1:]]
                return F.add_n(*means)
            else:
                return F.zeros(1)
        return F.zeros(1)
